from .curation import *

from .access import (
    SHClient,
    SHParametersFeature,
    EvalScripts,
    sentinel_1_download_parameters,
    sentinel_2_l2a_download_parameters,
    dem_download_parameters,
    AirbusClient,
    get_airbus_access_token,
)
from .tools import (
    stac_items_to_gdf,
    get_images_by_location,
    calculate_average_coordinates_distance,
    generate_new_locations_bounding_boxes,
    get_tarfile_image_info,
    get_first_last_dates,
    create_time_slots,
    expand_time_interval,
    generate_location_payload,
    format_product_location_payload,
)

__version__ = "2023.10.24-4"



'''
Module for EO STAC extensions object
'''

import pystac
import pandas as pd

from typing import Union
from .base import STACExtensionObject
from pystac.extensions.eo import Band, EOExtension


class EOS2ExtensionObject(STACExtensionObject):
    def __init__(self) -> None:
        super().__init__()
        self.bands_dict = {
            "B01": Band.create(
                name="B01",
                description="Coastal aerosol, 442.7 nm (S2A), 442.3 nm (S2B)",
                common_name="coastal",
            ),
            "B02": Band.create(
                name="B02",
                description="Blue, 492.4 nm (S2A), 492.1 nm (S2B)",
                common_name="blue",
            ),
            "B03": Band.create(
                name="B03",
                description="Green, 559.8 nm (S2A), 559.0 nm (S2B)",
                common_name="green",
            ),
            "B04": Band.create(
                name="B04",
                description="Red, 664.6 nm (S2A), 665.0 nm (S2B)",
                common_name="red",
            ),
            "B05": Band.create(
                name="B05",
                description="Vegetation red edge, 704.1 nm (S2A), 703.8 nm (S2B)",
                common_name="rededge",
            ),
            "B06": Band.create(
                name="B06",
                description="Vegetation red edge, 740.5 nm (S2A), 739.1 nm (S2B)",
                common_name="rededge",
            ),
            "B07": Band.create(
                name="B07",
                description="Vegetation red edge, 782.8 nm (S2A), 779.7 nm (S2B)",
                common_name="rededge",
            ),
            "B08": Band.create(
                name="B08",
                description="NIR, 832.8 nm (S2A), 833.0 nm (S2B)",
                common_name="nir",
            ),
            "B08a": Band.create(
                name="B08a",
                description="Narrow NIR, 864.7 nm (S2A), 864.0 nm (S2B)",
                common_name="nir08",
            ),
            "B09": Band.create(
                name="B09",
                description="Water vapour, 945.1 nm (S2A), 943.2 nm (S2B)",
                common_name="nir09",
            ),
            "B10": Band.create(
                name="B10",
                description="SWIR – Cirrus, 1373.5 nm (S2A), 1376.9 nm (S2B)",
                common_name="cirrus",
            ),
            "B11": Band.create(
                name="B11",
                description="SWIR, 1613.7 nm (S2A), 1610.4 nm (S2B)",
                common_name="swir16",
            ),
            "B12": Band.create(
                name="B12",
                description="SWIR, 2202.4 nm (S2A), 2185.7 nm (S2B)",
                common_name="swir22",
            ),
        }

    def add_extension_to_object(
        self, obj: Union[pystac.Item, pystac.Asset],
        obj_info: pd.DataFrame
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        # Add EO extension
        eo_ext = EOExtension.ext(obj, add_if_missing=True)
        # Add common metadata
        if isinstance(obj, pystac.Item) or (
            isinstance(obj, pystac.Asset) and obj.title not in self.bands_dict.keys()
        ):
            obj.common_metadata.constellation = "Sentinel-2"
            obj.common_metadata.platform = "Sentinel-2"
            obj.common_metadata.instruments = ["Sentinel-2"]
            obj.common_metadata.gsd = 10
            # Add bands
            bands = obj_info["bands"].values
            bands = bands[0] if bands else None
            bands_list = [self.bands_dict[band] for band in bands] if bands else None
            eo_ext.apply(bands=bands_list)

        elif isinstance(obj, pystac.Asset):
            eo_ext.apply(bands=[self.bands_dict[obj.title]])

        return obj



from .stac.dataframe import STACDataFrame  # , read_stac

from .stac.stac import STACGenerator
from .folder_formatters import SHFolderFormatter
from .stac.utils import format_time_acquired, merge_stac_catalogs
from .stac.parsers import STACIdParser, StructuredParser, UnestructuredParser
from .stac.dataframe_labeling import UnlabeledStrategy, LabeledStrategy



from .client import SHClient
from .parameters import (SHParametersFeature,
                        sentinel_1_download_parameters,
                        sentinel_2_l2a_download_parameters,
                        dem_download_parameters)
from .evalscripts import EvalScripts


"""
Module for STAC extensions objects
"""

from typing import Optional, Union
import pystac

import pandas as pd


class STACExtensionObject:
    def __init__(self) -> None:
        super().__init__()
        self.properties = dict()

    def add_extension_to_object(
        self, obj: Union[pystac.Item, pystac.Asset],
        obj_info: Optional[pd.DataFrame] = None
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        pass



from .DownloadDataset import DownloadDataset
from .DownloadFileURL import DownloadFileURL
from .IngestDataset import IngestDataset
from .IngestLargeDataset import IngestLargeDataset
from .RetrieveDataset import RetrieveDataset
from .RetrieveDatasets import RetrieveDatasets
from .IngestFile import IngestFile
from .IngestFolder import IngestFolder
from .IngestSTAC import IngestSTAC



"""
Module for data engineeringt
"""

import geopandas as gpd
import pandas as pd
import tarfile
import rasterio
import re
import datetime
import json

from shapely.geometry import box, Polygon
from pyproj import Transformer
from os.path import exists
from typing import Union, Optional


def get_images_by_location(gdf: gpd.GeoDataFrame) -> pd.DataFrame:
    """
    Generate a GeoDataFrame with the available images for each location in the dataset. 

    :param gdf: GeoDataFrame generated from the ItemCollection of a sen12floods collection
    :return gdf_dates_per_aoi:  GeoDataFrame with the available images for each location in
                                the dataset.
            The returned GeoDataFrame has three columns:
                - location_id: the unique ID of each location.
                - images_count: the count of available images of each location.
                - images_dates: list with the dates of the available images of each location.
    """
    uniques_location_id = gdf['scene_id'].unique()   # List of unique location ids
    uniques_location_id.sort()

    images_count_list, images_dates_list = [], []

    # Iterate the unique location ids, count the number of images per location and generate
    # a list with the dates of every image in a location
    for location_id in uniques_location_id:
        dates = gdf[gdf['scene_id'] == location_id]['datetime']
        images_count_list.append(dates.count())
        images_dates_list.append(dates.tolist())

    images_dates_list.sort()   # Sort the list of dates
    data = {'scene_id': uniques_location_id, 'dates_count': images_count_list, 'dates_list': images_dates_list}
    df_dates_per_aoi = pd.DataFrame.from_dict(data)

    return df_dates_per_aoi


def generate_location_payload(gdf: Union[gpd.GeoDataFrame, pd.DataFrame], path: str) -> dict:
    """
    Generate a dictionary with the location payload of the locations in the GeoDataFrame, 
    such as the bounding box and the time interval to search for available data.
    """
    payload_cache = f"{path}/location_payload.json"
    if exists(payload_cache):
        # Read as dict
        with open(payload_cache, 'r') as f:
            payload = json.load(f)
        return payload
    
    bbox_date_by_location = dict()
    for i, row in gdf.iterrows():
        # Get list from dates_list column
        dates_list = list(row['dates_list'])
        for date in dates_list:
            location_id = row['location_id']
            date_formatted = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%Y-%m-%d')
            location_id_formatted = f'{location_id}_{date_formatted}'
            bbox_date_by_location[location_id_formatted] = {
                'bounding_box': row['geometry'].bounds,
                # Convert str to datetime
                'time_interval': (date, date)
            }

    # Save to json
    with open(payload_cache, 'w') as f:
        json.dump(bbox_date_by_location, f)

    return bbox_date_by_location


def get_tarfile_image_info(tar: str, path: Optional[str] = None, pattern: Optional[str] = r"\d{8}T\d{6}", level: Optional[int] = 2):
    """
    """
    if path:
        gdf_cache = f"{path}/tarfile_images_info.csv"
        if exists(gdf_cache):
            images_gdf = gpd.read_file(gdf_cache,
                                       GEOM_POSSIBLE_NAMES="geometry", 
                                       KEEP_GEOM_COLUMNS="NO")
            images_gdf.set_crs(epsg=4326, inplace=True)
            
            return images_gdf
    
    images_df = pd.DataFrame()
    with tarfile.open(tar, 'r:gz') as tar:
        rasters = [i for i in tar.getnames() if i.endswith(".tif") or i.endswith(".tiff")]
        for raster in rasters:
            r = tar.extractfile(raster)
            bbox = get_image_bbox(r)
            date = extract_image_date_in_folder(raster, pattern)
            date_formatted = datetime.datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%Y-%m-%d')
            id = extract_image_id_in_folder(raster, level)
            # Use pd.concat to append to dataframe
            images_df = pd.concat([images_df, pd.DataFrame({"location_id": [id], 
                                                            "datetime": [date], 
                                                            "bbox": [bbox]})])

    # Clean duplicates
    images_df = images_df.drop_duplicates(subset=["location_id", "datetime"])
    # Convert to geodataframe
    images_gdf = gpd.GeoDataFrame(images_df, 
                                  crs='EPSG:4326',
                                  geometry=images_df["bbox"].apply(lambda x: box(x[0], x[1], x[2], x[3])))
    # Drop bbox column
    images_gdf = images_gdf.drop(columns=["bbox"])
    # Set crs
    images_gdf = images_gdf.set_crs(epsg=4326)
    # Sort by location_id
    images_gdf = images_gdf.sort_values(by=["location_id"])
    if path:
        # Save to csv
        images_gdf.to_csv(gdf_cache, index=False)

    return images_gdf


def get_image_bbox(raster: Union[tarfile.ExFileObject, str]):
    with rasterio.open(raster) as src:
        bbox = src.bounds
    return bbox


def get_image_resolution(raster: Union[tarfile.ExFileObject, str]):
    with rasterio.open(raster) as src:
        resolution = src.res
    return resolution


def extract_image_date_in_folder(raster_path: str, pattern: str):
    case = re.findall(pattern, raster_path)

    if case:
        date = case[0]
        # Convert date to format YYYY-MM-DDT00:00:00.000Z as datetime object
        formatted_date = datetime.datetime.strptime(date, '%Y%m%dT%H%M%S').strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return formatted_date
    
    return None


def extract_image_id_in_folder(raster_path: str, level: int):
    return raster_path.split("/")[level]


def get_first_last_dates(dataframe: Union[pd.DataFrame, gpd.GeoDataFrame], dates_column: Optional[str] = 'datetime'):
    """
    """
    dataframe[dates_column] = dataframe[dates_column].apply(lambda x: sorted(x))
    dataframe['first_date'] = dataframe['dates_list'].apply(lambda x: x[0])
    dataframe['last_date'] = dataframe['dates_list'].apply(lambda x: x[-1])
    dataframe = dataframe.sort_values(by=['first_date', 'last_date'])
    # Sort by sequence id
    dataframe = dataframe.sort_values(by=['location_id'])
    # Convert first_date and last_date to datetime, in format YYYY-MM-DD
    dataframe['first_date'] = pd.to_datetime(dataframe['first_date'], format='%Y-%m-%d')
    dataframe['last_date'] = pd.to_datetime(dataframe['last_date'], format='%Y-%m-%d')

    return dataframe


def create_time_slots(start_date: datetime.datetime, end_date: datetime.datetime, n_chunks: int):
    """
    """
    tdelta = (end_date - start_date) / n_chunks
    edges = [(start_date + i * tdelta).date().isoformat() for i in range(n_chunks)]
    slots = [(edges[i], edges[i + 1]) for i in range(len(edges) - 1)]

    return slots


def expand_time_interval(time_interval: Union[list, tuple], format: str='%Y-%m-%dT%H:%M:%S.%fZ') -> list:
    """
    """
    start_date = time_interval[0]
    end_date = time_interval[1]

    if isinstance(start_date, str):
        start_date = datetime.datetime.strptime(start_date, format)
    if isinstance(end_date, str):
        end_date = datetime.datetime.strptime(end_date, format)

    # Add one day to start date and remove one day to end date
    new_start_date = start_date - datetime.timedelta(days=1)
    new_end_date = end_date + datetime.timedelta(days=1)

    # Convert to string
    new_start_date = new_start_date.strftime(format)
    new_end_date = new_end_date.strftime(format)

    return new_start_date, new_end_date


def format_product_location_payload(location_payload: dict,
                                    images_response: dict,
                                    all_info: bool = False
                                    ) -> dict:
    """
    """
    for id, info in location_payload.items():
        # Add new key to the dictionary
        if all_info:
            location_payload[id]['image'] = images_response[id] if id in images_response else None
        else:
            location_payload[id]['image'] = images_response[id]['properties']['id'] if images_response[id] else None

    return location_payload


def bbox_to_coordinates(bounding_box: list) -> list:
    """
    """
    polygon_coordinates = [
        (bounding_box[0], bounding_box[1]),  # bottom left
        (bounding_box[0], bounding_box[3]),  # top left
        (bounding_box[2], bounding_box[3]),  # top right
        (bounding_box[2], bounding_box[1]),  # bottom right
        (bounding_box[0], bounding_box[1])   # back to bottom left
    ]

    return polygon_coordinates


def bbox_to_polygon(bounding_box: list) -> Polygon:
    """
    """
    polygon = box(bounding_box[0], bounding_box[1], bounding_box[2], bounding_box[3])

    return polygon


from_4326_transformer = Transformer.from_crs('EPSG:4326', 'EPSG:3857')
from_3857_transformer = Transformer.from_crs('EPSG:3857', 'EPSG:4326')


def bbox_from_centroid(x: Union[int, float], 
                       y: Union[int, float],
                       pixel_size: Union[int, float],
                       width: Union[int, float],
                       height: Union[int, float]
                       ) -> list:
    """
    Generate a bounding box from a centroid, pixel size and image dimensions.

    Params
    ------
    x: int or float
        x coordinate of the centroid
    y: int or float
        y coordinate of the centroid
    pixel_size: int or float
        pixel size in meters
    width: int or float
        width of the image in pixels
    height: int or float
        height of the image in pixels

    Returns
    -------
    bounding_box: list
        list with the bounding box coordinates
    """
    width_m = width * pixel_size
    heigth_m = height * pixel_size

    # Transform the centroid coordinates to meters
    centroid_m = from_4326_transformer.transform(x, y)

    # Calculate the bounding box coordinates
    min_x = centroid_m[0] - width_m / 2
    min_y = centroid_m[1] - heigth_m / 2
    max_x = centroid_m[0] + width_m / 2
    max_y = centroid_m[1] + heigth_m / 2

    # Convert the bounding box coordinates back to degrees
    min_x, min_y = from_3857_transformer.transform(min_x, min_y)
    max_x, max_y = from_3857_transformer.transform(max_x, max_y)

    return [min_y, min_x, max_y, max_x]



'''
Module for STAC Asset Generators
'''

from os import remove, listdir
from os.path import dirname, join, basename, abspath
from ..metadata import remove_raster_metadata
from typing import List

import pandas as pd
import rasterio
import pystac

MEDIA_TYPES_DICT = {
    'tif': pystac.MediaType.GEOTIFF,
    'tiff': pystac.MediaType.GEOTIFF,
    'png': pystac.MediaType.PNG,
    'jpg': pystac.MediaType.JPEG,
    'jpeg': pystac.MediaType.JPEG,
}


class STACAssetGenerator:

    type = 'None'
    
    def __init__(self):
        pass

    @classmethod
    def extract_assets(self, obj_info: pd.DataFrame):
        """
        Generate a single asset from the raster file

        :param raster_path: path to the raster file
        """
        # If there is no bands, create a single band asset from the file, assuming thats a singleband raster
        raster_path = obj_info["image"].values[0]
        title = basename(raster_path).split('.')[0]
        # Get the file extension
        raster_format = raster_path.split('.')[-1]
        asset = pystac.Asset(href=abspath(raster_path), 
                             title=title, 
                             media_type=MEDIA_TYPES_DICT[raster_format], 
                             roles=['data'])

        return [asset]


class BandsAssetGenerator(STACAssetGenerator):

    type = 'Bands'

    def __init__(self) -> None:
        super().__init__()
    
    def extract_assets(self, obj_info: pd.DataFrame):
        """
        Extract the assets from the raster file from the bands column

        :param raster_path: path to the raster file
        """
        asset_list = []
        # File path
        raster_path = obj_info["image"].values[0]
        # Bands
        bands = obj_info["bands"].values
        bands = bands[0] if bands else None

        if bands:
            with rasterio.open(raster_path, 'r') as raster:
                if isinstance(bands, str):
                    bands = [bands]
                for band in bands:
                    i = bands.index(band)
                    raster_format = raster_path.split('.')[-1]   # Will be used later to save the bands files
                    try:
                        single_band = raster.read(i + 1)
                    except IndexError:
                        single_band = raster.read(1)
                    band_name = f'{band}.{raster_format}'
                    output_band = join(dirname(raster_path), band_name)
                    # Copy the metadata
                    metadata = raster.meta.copy()
                    metadata.update({"count": 1})
                    # Write the band to the output folder
                    with rasterio.open(output_band, "w", **metadata) as dest:
                        dest.write(single_band, 1)
                    # Instantiate pystac asset and append it to the list
                    asset_list.append(pystac.Asset(href=output_band, 
                                                   title=band, 
                                                   media_type=MEDIA_TYPES_DICT[raster_format]))

            # Remove the original raster file and its metadata
            remove(raster_path)
            remove_raster_metadata(dirname(raster_path))

            return asset_list


class ExtractedAssets(STACAssetGenerator):

    type = 'Extracted'

    def __init__(self) -> None:
        super().__init__()

    def extract_assets(self, obj_info: pd.DataFrame):
        """
        Get all the files with the same extension as the image file as assets
        """
        asset_list = []
        # File path
        raster_path = obj_info["image"].values[0]
        raster_dir = dirname(raster_path)
        # Get the files with the same extension as the image file
        files = [f for f in listdir(raster_dir) if f.endswith(raster_path.split('.')[-1])]
        # Instantiate pystac asset and append it to the list
        for file in files:
            # Get the file extension
            raster_format = file.split('.')[-1]
            asset_list.append(pystac.Asset(href=join(raster_dir, file), 
                                           title=basename(file), 
                                           media_type=MEDIA_TYPES_DICT[raster_format]))

        return asset_list



"""
Utils
"""

import requests


def get_airbus_access_token(api_key: str) -> str:
  """
  Get Airbus access token

  Returns:
    str: access token
  """
  headers = {
      'Content-Type': 'application/x-www-form-urlencoded',
  }

  data = [
    ('apikey', api_key),
    ('grant_type', 'api_key'),
    ('client_id', 'IDP'),
  ]

  response = requests.post('https://authenticate.foundation.api.oneatlas.airbus.com/auth/realms/IDP/protocol/openid-connect/token', headers=headers, data=data)

  access_token = response.json()['access_token']

  return access_token



"""
Module for data engineering with STAC elements
"""
import geopandas as gpd
from pystac import ItemCollection


def stac_items_to_gdf(items: ItemCollection) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame from a given pystac.ItemCollection. 

    :param: items: A pystac.ItemCollection
    :return: GeoDataframe from the given ItemCollection
    """
    _features = [i.to_dict() for i in items]

    # Get a new ItemCollection by removing duplicate items, if they exist
    features = []
    for f in _features:
        if f not in features:
            # Add the id, type and stac_extensions as properties, in order to retrieve 
            # them as columns in the GeoDataFrame
            # TODO put all the properties in the GeoDataFrame
            f['properties']['id'] = f['id']
            if 'scene_id' in f['properties']:
                f['properties']['scene_id'] = f['id'].split('_')[3]
            f['properties']['type'] = f['type']
            f['properties']['stac_extensions'] = f['stac_extensions']
            features.append(f)       
    
    return gpd.GeoDataFrame.from_features(features)



'''
Module for raster STAC extensions object
'''

import pystac
import rasterio
import pandas as pd

from pystac.extensions.raster import RasterExtension, RasterBand
from typing import Union, Optional
from .base import STACExtensionObject


class RasterExtensionObject(STACExtensionObject):
    def __init__(self) -> None:
        super().__init__()

    def add_extension_to_object(
        self, obj: Union[pystac.Item, pystac.Asset],
        obj_info: Optional[pd.DataFrame] = None
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        if not isinstance(obj, pystac.Asset):
            return obj
        else:
            raster_ext = RasterExtension.ext(obj, add_if_missing=True)
            src = rasterio.open(obj.href)
            bands = list()
            for band in src.indexes:
                bands.append(RasterBand.create(
                    nodata=src.nodatavals[band - 1],
                    data_type=src.dtypes[band - 1],
                    spatial_resolution=src.res) if src.nodatavals else RasterBand.create(
                        data_type=src.dtypes[band - 1],
                        spatial_resolution=src.res))
            raster_ext.apply(bands=bands)
                
        return obj





'''
STAC utils
'''

import pystac

from os.path import dirname, join, abspath
from typing import Union, Optional
from tqdm import tqdm
from traceback import print_exc
from shutil import rmtree


def get_all_children(obj: pystac.STACObject) -> list:
    """
    Get all the children of a STAC object

    :param obj: STAC object
    """
    children = []
    # Append the current object to the list
    children.append(obj.to_dict())

    # Collections
    collections = list(obj.get_collections())

    for collection in collections:
        children.append(collection.to_dict())

    # Items
    items = obj.get_items()
    for item in items:
        children.append(item.to_dict())

    # Items from collections
    for collection in collections:
        items = collection.get_items()
        for item in items:
            children.append(item.to_dict())

    return children


def make_links_relative_to_path(path: str,
                                catalog: Union[pystac.Catalog, str],
                                ) -> pystac.Catalog:
    """
    Makes all asset HREFs in the catalog relative to a given path

    :param path: path to make the links relative
    :param catalog: catalog to make the links relative

    :return: catalog with the links relative
    """
    if isinstance(catalog, str):
        catalog = pystac.read_file(catalog)
    path = abspath(path)

    catalog.make_all_asset_hrefs_absolute()

    for collection in catalog.get_children():
        new_collection = collection.clone()
        new_collection.set_self_href(join(path, collection.id, f"collection.json"))
        new_collection.set_root(catalog)
        new_collection.set_parent(catalog)
        catalog.remove_child(collection.id)
        catalog.add_child(new_collection)
        for item in collection.get_all_items():
            new_item = item.clone()
            new_item.set_self_href(join(path, item.id, f"{item.id}.json"))
            new_item.set_parent(collection)
            new_item.set_root(catalog)
            new_collection.add_item(new_item)

    catalog.make_all_asset_hrefs_relative()

    return catalog


def merge_stac_catalogs(catalog_1: Union[pystac.Catalog, str],
                        catalog_2: Union[pystac.Catalog, str],
                        destination: Optional[str] = None,
                        keep_extensions: Optional[bool] = False,
                        catalog_type: Optional[pystac.CatalogType] = pystac.CatalogType.SELF_CONTAINED
                        ) -> None:
    """
    Merge two STAC catalogs, keeping the properties, collection and items of both catalogs

    :param catalog_1: first catalog to merge
    :param catalog_2: second catalog to merge
    :param destination: destination folder to save the merged catalog
    :param keep_extensions: keep the extensions of the first catalog
    :param catalog_type: type of the catalog
    """
    if isinstance(catalog_1, str):
        catalog_1 = pystac.Catalog.from_file(catalog_1)
    if isinstance(catalog_2, str):
        catalog_2 = pystac.Catalog.from_file(catalog_2)

    for col1 in tqdm(catalog_1.get_children(), desc='Merging catalogs...'):
        # Check if the collection exists in catalog_2
        col2 = catalog_2.get_child(col1.id)
        if col2 is None:
            # If it does not exist, add it
            col1_ = col1.clone()
            catalog_2.add_child(col1)
            col2 = catalog_2.get_child(col1.id)
            col2.clear_items()
            for i in col1_.get_stac_objects(pystac.RelType.ITEM):
                col2.add_item(i)
        else:
            # If it exists, merge the items
            for item1 in col1.get_items():
                if col2.get_item(item1.id) is None:
                    col2.add_item(item1)

    if keep_extensions:
        for ext in catalog_1.stac_extensions:
            if ext not in catalog_2.stac_extensions:
                catalog_2.stac_extensions.append(ext)

        for extra_field_name, extra_field_value in catalog_1.extra_fields.items():
            if extra_field_name not in catalog_2.extra_fields:
                catalog_2.extra_fields[extra_field_name] = extra_field_value

    if destination:
        # TODO test
        make_links_relative_to_path(destination, catalog_2)
    else:
        destination = dirname(catalog_2.get_self_href())

    # Save the merged catalog
    try:
        print("Validating and saving...")
        catalog_2.validate()
        rmtree(destination) if not destination else None # Remove the old catalog and replace it with the new one
        catalog_2.normalize_and_save(root_href=destination, 
                                     catalog_type=catalog_type
                                     )
        print("Success!")
    except pystac.STACValidationError:
        # Return full callback
        print_exc()



"""
Utils
"""

from sentinelhub import DataCollection, MosaickingOrder

from .evalscripts import EvalScripts
from ..parameters import ParametersFeature


class SHParametersFeature(ParametersFeature):
    """
    Class that allows the control of the parameters needed to search
    or download data from Sentinel Hub
    """

    def __init__(self, parameters: dict) -> None:
        """
        :param parameters: dict that contains the parameters that are going to be used to search or download Sentinel Hub data.
                The dictionary could contain the following parameters in the following format and names:
                    - data_collection: sentinelhub.DataCollection object with the desired data collection to search or download.
                    - data_to_download: dict with the unique ID and bounding box of every location to request sentinel data from.
                    - bbox: list with the bounding box of the location.
                    - time_interval: list with the time interval stablished.
                    - resolution: resolution of the image to download.
                    - data_folder: location of the directory where the fetched data will be saved.
                    - fields: dict with the fields desired to include or exclude. It must follow the format:
                        {
                        "include": list of fields to include, 
                        "exclude": list of fields to exclude
                        }
                    - filter: string with the filter to apply, if needed.
                    - evalscript: is a piece of Javascript code which defines how the satellite data shall be 
                    processed by Sentinel Hub and what values the service shall return. For further information,
                    see https://docs.sentinel-hub.com/api/latest/evalscript/
                    - mosaicking_order: sentinelhub.MosaickingOrder object with the desired mosaicking order
        """
        ParametersFeature.__init__(self, parameters)
        self.data_collection = parameters['data_collection'] if 'data_collection' in parameters else None
        self.data_to_download = parameters['data_to_download'] if 'data_to_download' in parameters else None
        self.bounding_box = parameters['bounding_box'] if 'bounding_box' in parameters else None
        self.time_interval = parameters['time_interval'] if 'time_interval' in parameters else None
        self.fields = parameters['fields'] if 'fields' in parameters else None
        self.filter = parameters['filter'] if 'filter' in parameters else None
        self.evalscript = parameters['evalscript'] if 'evalscript' in parameters else None
        self.resolution = parameters['resolution'] if 'resolution' in parameters else None
        self.data_folder = parameters['data_folder'] if 'data_folder' in parameters else None
        self.mosaicking_order = parameters['mosaicking_order'] if 'mosaicking_order' in parameters else None


options = {
        'data_collection': DataCollection.SENTINEL1,
        'fields': {"include": ["id", 
                        "properties.datetime",
                        "sar:instrument_mode", 
                        "s1:polarization",
                        "sat:orbit_state",
                        "s1:resolution",
                        "s1:timeliness"], 
            "exclude": []},
        'filter': None
    }
sentinel_1_search_parameters = SHParametersFeature(options)


options = {
        'data_collection': DataCollection.SENTINEL2_L2A,
        'fields': {"include": ["id", 
                        "properties.datetime"], 
            "exclude": []}
    }
sentinel_2_l2a_search_parameters = SHParametersFeature(options)


options = {
    'data_collection': DataCollection.SENTINEL1,
    'resolution': 3,
    'evalscript': EvalScripts.SENTINEL_1
}
sentinel_1_download_parameters = SHParametersFeature(options)


options = {
    'data_collection': DataCollection.SENTINEL2_L2A,
    'resolution': (10, 10),
    'mosaicking_order': MosaickingOrder.LEAST_CC,
    'evalscript': EvalScripts.SENTINEL_2
}
sentinel_2_l2a_download_parameters = SHParametersFeature(options)


options = {
    'data_collection': DataCollection.DEM_COPERNICUS_30,
    'resolution': 3,
    'evalscript': EvalScripts.DEM,
}
dem_download_parameters = SHParametersFeature(options)



'''
Module for SAR STAC extensions object
'''

import pystac
import pandas as pd

from .base import STACExtensionObject

from typing import Optional, Union
from pystac.extensions.sar import SarExtension
from pystac.extensions.sar import FrequencyBand, Polarization


class SarExtensionObject(STACExtensionObject):
    def __init__(self) -> None:
        super().__init__()
        self.polarizations = [Polarization.VV, Polarization.VH]
        self.polarizations_dict = {"VV": Polarization.VV, "VH": Polarization.VH}

    def add_extension_to_object(
        self, obj: Union[pystac.Item, pystac.Asset],
        obj_info: Optional[pd.DataFrame] = None
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        # Add SAR extension to the item
        sar_ext = SarExtension.ext(obj, add_if_missing=True)
        if isinstance(obj, pystac.Item) or (
            isinstance(obj, pystac.Asset)
            and obj.title not in self.polarizations_dict.keys()
        ):
            polarizations = self.polarizations
        elif (
            isinstance(obj, pystac.Asset)
            and obj.title in self.polarizations_dict.keys()
        ):
            polarizations = [self.polarizations_dict[obj.title]]
        sar_ext.apply(
            instrument_mode="EW",
            polarizations=polarizations,
            frequency_band=FrequencyBand.C,
            product_type="GRD",
        )

        return obj



'''
Module for the STAC label extension ScaneoLabeler object
'''

import pystac
import pandas as pd
import json

from glob import glob
from tqdm import tqdm
from os.path import join, dirname, exists, splitext, basename, abspath
from typing import List, Optional, Union
from pystac.extensions.label import LabelExtension

from .base import LabelExtensionObject
from ...extent import get_unknow_extent


class ScaneoLabeler(LabelExtensionObject):
    def __init__(self) -> None:
        super().__init__()

    def generate_stac_labels(
        self,
        catalog: Union[pystac.Catalog, str],
        root_folder: str,
        collection: Optional[Union[pystac.Collection, str]] = 'source',
        label_description: Optional[str] = "Item label",
        label_type: Optional[str] = "vector",
        label_names: Optional[List[str]] = ["label"],
        **kwargs
    ) -> None:
        """
        Generate a labels collection from a STAC dataframe. 
        This class should be used when the items have been labeled using SCANEO, as is implemented
        taking into account the SCANEO labeling format.
        
        :param catalog: catalog to add the labels collection to
        :param root_folder: root folder where are the images and the labels as GeoJSON files, following the SCANEO labeling format
        :param stac_dataframe: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param collection: collection to add the labels collection to
        :param label_description: label description
        :param label_type: label type
        :param label_names: list of label names
        :param kwargs: optional arguments
            :param kwargs.label_properties: list of label properties
            :param kwargs.label_methods: list of label methods
        """
        if isinstance(catalog, str):
            catalog = pystac.Catalog.from_file(catalog)

        # Add the labels collection to the catalog
        # If exists a source collection, get it extent
        source_collection = catalog.get_child(collection)
        if source_collection:
            extent = source_collection.extent
            source_items = source_collection.get_stac_objects(pystac.RelType.ITEM)
        else:
            if not collection:
                raise ValueError(
                    "No source collection provided, please provide a source collection"
                )
            extent = get_unknow_extent()

        # Create the labels collection and add it to the catalog if it does not exist
        # If it exists, remove it
        collection = pystac.Collection(id="labels", description="Labels", extent=extent)
        if collection.id in [c.id for c in catalog.get_children()]:
            catalog.remove_child(collection.id)
        catalog.add_child(collection)

        # Get the GeoJSON files
        geojson_files = glob(join(root_folder, "*.geojson"))
        if not geojson_files:
            raise ValueError(
                "No GeoJSON files found in the root folder, please provide a root folder with the GeoJSON files"
            )

        # Get the label classes
        label_classes = self.get_label_classes(root_folder, geojson_files)

        # Generate the labels items
        for source_item in tqdm(source_items, desc="Generating labels collection..."):
            # Get the GeoJSON label of the item
            geojson_label = self.get_geojson_of_item(source_item, geojson_files)
            # Get the tasks from the GeoJSON label
            tasks = self.get_tasks_from_geojson(geojson_label)
            # Add the tasks to the kwargs
            kwargs['label_tasks'] = tasks

            # Create the label item
            label_item = self.add_extension_to_item(
                source_item,
                label_description=label_description,
                label_type=label_type,
                label_names=label_names,
                label_classes=label_classes,
                **kwargs
            )
            # Add the self href to the label item, following the Best Practices Layout
            # https://github.com/radiantearth/stac-spec/blob/master/best-practices.md
            label_item.set_self_href(
                join(
                    dirname(collection.get_self_href()),
                    label_item.id,
                    f"{label_item.id}.json"
                    )
            )
            # Match the GeoJSON label with the label item
            self.add_geojson_to_item(label_item, geojson_label, label_type)
            # Add the item to the collection
            collection.add_item(label_item)

        # Add the extension to the collection
        self.add_extension_to_collection(
            collection,
            label_names=[label_names],
            label_classes=[label_classes],
            label_type=label_type,
        )

        # Validate and save the catalog
        # Before adding the geojson, we need to save the catalog
        # and then iterate over the items to add the geojson
        try:
            pystac.validation.validate(catalog)
            catalog.normalize_and_save(dirname(catalog.get_self_href()), pystac.CatalogType.SELF_CONTAINED)
            print('Success on labels generation!')
        except pystac.STACValidationError as e:
            raise pystac.STACError(f"Catalog validation error: {e}")

    def add_geojson_to_item(self, 
                            item: pystac.Item,
                            geojson_path: str,
                            label_type: str
                            ) -> None:
        """
        Add a GeoJSON FeatureCollection to every label item, as recommended by the spec
        https://github.com/stac-extensions/label#assets

        :param collection: collection to add the labels collection to
        :param df: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param label_type: label type
        """
        properties = {'roles': ['labels', f'labels-{label_type}']}

        label_ext = LabelExtension.ext(item, add_if_missing=True)
        item.make_asset_hrefs_absolute()
        label_ext.add_geojson_labels(href=abspath(geojson_path), 
                                     title='Label', 
                                     properties=properties)
        item.make_asset_hrefs_relative()
        
    def get_label_classes(self,
                          root_folder: str,
                          geojsons: List[str]
                          ) -> List[str]:
        """
        Get the label classes from the labels.json file if exists, or from the GeoJSON files instead
        """
        label_classes = list()

        labels_json = glob(join(root_folder, "labels.json"))[0]
        if exists(labels_json):
            with open(labels_json, 'r') as f:
                labels = json.load(f)
            for value in labels['labels']:
                label_classes.append(value['name']) if value['name'] not in label_classes else None
        else:
            for geojson in geojsons:
                with open(geojson, 'r') as f:
                    labels = json.load(f)
                for value in labels['features']:
                    label_classes.append(value['properties']['labels']) if value['properties']['labels'] not in label_classes else None

        return [label_classes]

    def get_geojson_of_item(self,
                            item: pystac.Item,
                            geojsons: List[str]
                            ) -> str:
        """
        Get the GeoJSON label of the item from a list of GeoJSON files

        :param item: item to get the GeoJSON label
        :param geojsons: list of GeoJSON files

        :return: path to the GeoJSON label of the item
        """
        item_id = item.id
        # Get a dict with <geojson_filename>: <geojson_path>, as the geojson_filename
        # must match the item ID
        geojsons_dict = dict(zip([splitext(basename(geojson))[0] for geojson in geojsons], geojsons))
        geojson_path = geojsons_dict.get(item_id)
        
        return geojson_path

    def get_tasks_from_geojson(self, 
                               geojson_path: str
                               ) -> List[str]:
        """
        Get the tasks from the GeoJSON label

        :param geojson_path: path to the GeoJSON label

        :return: list of tasks
        """
        with open(geojson_path, 'r') as f:
            geojson = json.load(f)
            tasks = list()
            for feature in geojson['features']:
                for task in feature['properties']['tasks']:
                    tasks.append(task) if task not in tasks else None
        
        return tasks



from pydantic import BaseModel


class DownloadFileURL:
    def __init__(self, repo, logger, progress=True):
        self.repo = repo
        self.logger = logger if logger else print
        self.progress = progress

    class Inputs(BaseModel):
        url: str
        path: str = None
        user: dict

    class Outputs(BaseModel):
        dst_path: str

    def __call__(self, inputs: Inputs) -> Outputs:
        dst_path = self.repo.download_file_url(
            inputs.url, inputs.path, inputs.user["id_token"], progress=self.progress
        )
        return self.Outputs(dst_path=dst_path)



import requests
import os
from .APIRepo import APIRepo


class AuthAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def login(self):
        return requests.get(self.url + "auth/login")

    def token(self, code):
        return requests.get(self.url + "auth/token?code=" + code)

    def logout_url(self):
        response = requests.get(self.url + "auth/logout")
        return response.json()["logout_url"]

    def retrieve_credentials(self, id_token):
        response = requests.get(
            self.url + "auth/credentials",
            headers={"Authorization": "Bearer " + id_token},
        )
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]



'''
Module for folder formatter classes
'''


class FolderFormatter:
    
    def format_folders(self) -> None:
        """
        Format the folder structure of the downloaded data to a 
        digestible format
        """
        pass

    def generate_raster_metadata(self) -> None:
        """
        Generate metadata.json file for a raster file
        """
        pass



import typer
from pathlib import Path

from ..datasets import (
    retrieve_datasets,
    ingest_dataset,
    download_dataset,
)

app = typer.Typer()


@app.command()
def list(
    name: str = typer.Option(None, "--name", "-n", help="Filter by name"),
    limit: int = typer.Option(None, "--limit", "-l", help="Limit number of results"),
):
    try:
        datasets = retrieve_datasets(name, limit)
        typer.echo(datasets)
    except Exception as e:
        typer.echo(e)


@app.command()
def ingest(
    path: Path = typer.Option(..., "--path", "-p", help="Path to dataset"),
    verbose: bool = typer.Option(False, "--verbose", help="Verbose output"),
):
    try:
        ingest_dataset(path, verbose, typer.echo)
    except Exception as e:
        typer.echo(e)


@app.command()
def get(
    dataset: str,
    path: str = typer.Option(None, "--path", "-p", help="Download to a specific path"),
    file: str = typer.Option(None, "--file", "-f", help="Download a specific file"),
    version: int = typer.Option(None, "--version", "-v", help="Dataset version"),
    assets: bool = typer.Option(False, "--assets", "-a", help="Download assets"),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force download even if file exists"
    ),
    verbose: bool = typer.Option(False, "--verbose", help="Verbose output"),
):
    try:
        dst_path = download_dataset(
            dataset, version, path, file, typer.echo, assets, force, verbose
        )
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


if __name__ == "__main__":
    app()



'''
Module for the labeling strategy when creating a STAC catalog from a dataframe
'''

import pystac
from os.path import basename


class LabelingStrategy:
    """
    Labeling strategy interface to be implemented by concrete labeling strategies
    """

    def __init__(self):
        pass

    def get_images_labels(self, images):
        pass


class UnlabeledStrategy(LabelingStrategy):
    """
    Assumes the images are not labeled, and returns the entire filename as label
    """

    def __init__(self):
        super().__init__()

    def get_images_labels(self, images):
        """
        """
        labels = list()
        for image in images:
            labels.append(basename(image).split('.')[0])
        ixs = [labels.index(x) for x in labels]

        return labels, ixs


class LabeledStrategy(LabelingStrategy):
    """
    Assumes the images are already labeled, and returns the labels.
    The images filenames must follow the pattern: <label>_<id>.<ext> or <label>-<id>.<ext>
    """

    def __init__(self):
        super().__init__()

    def get_images_labels(self, images):
        """
        """
        labels = list()
        for image in images:
            image_basename = basename(image).split('.')[0]   # Get filename without extension
            if '_' in image_basename:
                separator = '_'
            elif '-' in image_basename:
                separator = '-'
            try:
                # Ensure the image has format <label>_<number> or <label>-<number>
                int(image_basename.split(separator)[1])
                label = image_basename.split(separator)[0]
            except ValueError:
                label = image_basename
            labels.append(label)

        ixs = [labels.index(x) for x in labels]

        return labels, ixs



"""
Module for parameters features
"""

class ParametersFeature:
    """
    Class that allows the control of the parameters needed to perform 
    any data search or download
    """
    
    def __init__(self, parameters: dict) -> None:
        """
        :param parameters: dict that contains the parameters that are going to be used
        """
        pass



from .stac import stac_items_to_gdf
from .sen12floods import (calculate_average_coordinates_distance,
                           generate_new_locations_bounding_boxes)
from .tools import (generate_location_payload,
                    format_product_location_payload,
                    get_images_by_location,
                    get_tarfile_image_info,
                    get_first_last_dates,
                    create_time_slots,
                    expand_time_interval)



"""
Module for generating STAC metadata 
"""

import traceback
from typing import Union
import pandas as pd
import pystac
from tqdm import tqdm

from os.path import join, basename, dirname
from shutil import rmtree

import rasterio
import random
from rasterio.warp import transform_bounds
from typing import Union, List

from datetime import datetime
from shapely.geometry import Polygon, mapping
from glob import glob
from typing import Union, Optional

from .parsers import STACIdParser, StructuredParser
from .assets import STACAssetGenerator
from .dataframe_labeling import LabelingStrategy, UnlabeledStrategy
from .utils import (format_time_acquired, 
                    cut_images, 
                    get_item_metadata,
                    get_all_images_in_path)
from .extensions import (type_stac_extensions_dict, 
                         SUPPORTED_EXTENSIONS, 
                         LabelExtensionObject)
from .extent import (get_unknow_extent, 
                     get_collection_extent)


class STACGenerator:
    def __init__(
        self,
        image_format: str = "tiff",
        catalog_type: pystac.CatalogType = pystac.CatalogType.SELF_CONTAINED,
        item_parser: STACIdParser = StructuredParser,
        assets_generator: STACAssetGenerator = STACAssetGenerator,
        labeling_strategy: LabelingStrategy = UnlabeledStrategy,
    ) -> None:
        """
        Initialize the STAC generator

        :param image_format: image format of the assets
        :param catalog_type: type of the catalog
        :param item_parser: parser to get the item ID
        :param assets_generator: generator to generate the assets
        :param labeling_strategy: strategy to label the images
        """
        self._image_format = image_format
        self._catalog_type = catalog_type
        self._item_parser = item_parser()
        self._assets_generator = assets_generator()
        self._labeling_strategy = labeling_strategy()
        self._extensions_dict: dict = type_stac_extensions_dict
        self._stac_dataframe = pd.DataFrame()

    def generate_stac_metadata(
        self,
        id: str,
        description: str,
        stac_dataframe: pd.DataFrame = None,
        output_folder: str = "stac",
        kwargs: dict = {},
    ) -> None:
        """
        Generate STAC metadata for a given directory containing the assets to generate metadata

        :param id: id of the catalog
        :param description: description of the catalog
        :param stac_dataframe: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param output_folder: output folder to write the catalog to
        """
        self._stac_dataframe = (
            stac_dataframe if self._stac_dataframe.empty else self._stac_dataframe
        )
        if self._stac_dataframe.empty:
            raise ValueError("No STAC dataframe provided")

        # Create an empty catalog
        catalog = pystac.Catalog(id=id, description=description, **kwargs)
        
        # Add the collections to the catalog
        collections = self._stac_dataframe.collection.unique()
        for collection_path in collections:
            # Generate the collection
            collection = self.generate_stac_collection(collection_path)
            # Add the collection to the catalog
            catalog.add_child(collection)

        # Check there have been generate all the items from the images
        items_count = 0
        for collection in catalog.get_children():
            items = list(set([item.id for item in collection.get_items(recursive=True)]))
            items_count += len(items)
        if len(self._stac_dataframe) != items_count:
            raise pystac.STACError(
                "Not all the STAC items have been generated, please check the Item parser or the STAC dataframe. If you are using the StructuredParser, check that the images are in the correct folder structure."
            )

        # Add the catalog to the root directory
        catalog.normalize_hrefs(output_folder)

        # Validate the catalog
        print("Validating and saving catalog...")
        try:
            pystac.validation.validate(catalog)
            catalog.save(catalog_type=self._catalog_type)
            print("Success!")
        except pystac.STACValidationError as e:
            print(f"Catalog validation error: {e}")
            return

    def get_stac_dataframe(self, 
                           path: str, 
                           collections: Optional[Union[str, dict]]='source',
                           bands: Optional[dict]=None, 
                           extensions: Optional[dict]=None,
                           sample: Optional[int]=None
                           ) -> pd.DataFrame:
        """
        Get a dataframe with the STAC metadata of a given directory containing the assets to generate metadata

        :param path: path to the root directory
        :param collections: dictionary with the collections
        :param bands: dictionary with the bands
        :param extensions: dictionary with the extensions
        """
        images = get_all_images_in_path(path, self._image_format)
        if len(images) == 0:
            raise ValueError("No images found in the given path with the given extension. Please check the path and the extension")
        
        if self._assets_generator.type == 'Extracted':
            images = cut_images(images)

        if sample:
            try:
                images = random.sample(images, sample)
            except ValueError:
                raise ValueError(f"Sample size must be smaller than the number of images ({len(images)}). May be there are no images found in the given path with the given extension")

        labels, ixs = self._labeling_strategy.get_images_labels(images)
        bands_values = self._get_items_list_from_dict(labels, bands)
        extensions_values = self._get_items_list_from_dict(labels, extensions)

        if collections == "source":
            # List of path with the same value repeated as many times as the number of images
            collections_values = [join(path, "source") for i in range(len(images))]
        elif collections == '*':
            collections_values = [join(path, basename(dirname(image))) for image in images]
        else:
            try:
                collections_values = [join(path, value) for value in self._get_items_list_from_dict(labels, collections)]
            except TypeError:
                raise pystac.STACError('There is an error generating the collections. Please check the collections dictionary')

        df = pd.DataFrame({'image': images, 
                           'label': labels, 
                           'ix': ixs, 
                           'collection': collections_values, 
                           'extensions': extensions_values, 
                           'bands': bands_values
                           })
        
        self._stac_dataframe = df

        return df

    def _get_items_list_from_dict(self, labels: list, items: dict) -> list:
        """
        Get a list of items from a dictionary

        :param labels: list of labels
        :param items: dictionary with the items
        """
        if not items:
            # Create list of None with the same length as the labels list
            return [None for _ in labels]
        items_list = list()
        for label in labels:
            if label in items.keys():
                items_list.append(items[label])
            else:
                items_list.append(None)

        return items_list

    def generate_stac_collection(self, collection_path: str) -> pystac.Collection:
        """
        Generate a STAC collection from a directory containing the assets to generate metadata

        :param collection_path: path to the collection
        """
        # Get the images of the collection, as they are needed to obtain the collection extent
        collection_images = self._stac_dataframe[
            self._stac_dataframe["collection"] == collection_path
        ]["image"]
        # Get the collection extent
        extent = get_collection_extent(collection_images)
        # Create the collection
        collection_id = basename(collection_path)
        collection = pystac.Collection(
            id=collection_id, description="Collection", extent=extent
        )

        print(f"Generating {collection_id} collection...")
        for image in tqdm(collection_images):
            # Create the item
            item = self.create_stac_item(image)
            # Add the item to the collection
            collection.add_item(item)

        # Return the collection
        return collection

    def create_stac_item(self, raster_path: str, kwargs: dict = {}) -> pystac.Item:
        """
        Create a STAC item from a directory containing the raster files and the metadata.json file

        :param raster_path: path to the raster file
        """
        # Check if there is any metadata file in the directory associated to the raster file
        metadata = get_item_metadata(raster_path)

        # Obtain the bounding box from the raster
        with rasterio.open(raster_path) as ds:
            bounds = ds.bounds
            dst_crs = "EPSG:4326"
            try:
                left, bottom, right, top = rasterio.warp.transform_bounds(
                    ds.crs, dst_crs, *bounds
                )
            except rasterio.errors.CRSError:
                # If the raster has no crs, set the bounding box to 0
                left, bottom, right, top = 0, 0, 0, 0

        # Create bbox
        bbox = [left, bottom, right, top]

        # Create geojson feature
        # If the bounding box has no values, set the geometry to None
        geom = mapping(
            Polygon([[left, bottom], [left, top], [right, top], [right, bottom]])
        )

        # Initialize pySTAC item parameters
        params = dict()
        params["properties"] = dict()

        # Obtain the date acquired
        start_time, end_time = None, None
        if metadata and metadata["acquisition-date"] and metadata["type"] not in ('dem', 'DEM'):
            time_acquired = format_time_acquired(metadata["acquisition-date"])
        else:
            # Check if the type of the data is DEM
            if metadata and metadata["type"] and metadata["type"] in ("dem", "DEM"):
                time_acquired = None
                start_time = datetime.strptime("2011-01-01", "%Y-%m-%d")
                end_time = datetime.strptime("2015-01-07", "%Y-%m-%d")
                params["start_datetime"] = start_time
                params["end_datetime"] = end_time
            else:
                # Set unknown date
                time_acquired = datetime.strptime("2000-01-01", "%Y-%m-%d")

        # Obtain the item ID. The approach depends on the item parser
        id = self._item_parser.get_item_id(raster_path)
        # Add the item ID to the dataframe, to be able to get it later
        self._stac_dataframe.loc[
            self._stac_dataframe["image"] == raster_path, "id"
        ] = id
        
        # Instantiate pystac item
        item = pystac.Item(
            id=id, geometry=geom, bbox=bbox, datetime=time_acquired, **params
        )

        # Get the item info, from the raster path
        item_info = self._stac_dataframe[self._stac_dataframe["image"] == raster_path]
        # Get the extensions of the item
        extensions = item_info["extensions"].values
        extensions = extensions[0] if extensions else None

        # Add the required extensions to the item
        if extensions:
            if isinstance(extensions, str):
                extensions = [extensions]
            for extension in extensions:
                if extension not in SUPPORTED_EXTENSIONS:
                    raise ValueError(f"Extension {extension} not supported")
                else:
                    extension_obj = self._extensions_dict[extension]
                    extension_obj.add_extension_to_object(item, item_info)

        # Add the assets to the item
        assets = self._assets_generator.extract_assets(item_info)
        if not assets:
            # If there are not assets using the selected generator, try with the default
            assets = STACAssetGenerator.extract_assets(item_info)

        # Add the assets to the item
        if assets:
            for asset in assets:
                if isinstance(asset, pystac.Asset):
                    item.add_asset(asset.title, asset)
                    # Add the required extensions to the asset if required
                    if extensions:
                        if isinstance(extensions, str):
                            extensions = [extensions]
                        for extension in extensions:
                            if extension not in SUPPORTED_EXTENSIONS:
                                raise ValueError(f"Extension {extension} not supported")
                            else:
                                extension_obj = self._extensions_dict[extension]
                                extension_obj.add_extension_to_object(asset, item_info)
        item.set_self_href(join(dirname(raster_path), f"{id}.json"))
        item.make_asset_hrefs_relative()
        return item



'''
Time utils
'''

from typing import Union
from datetime import datetime
from dateutil import parser


def format_time_acquired(dt: Union[str, datetime]) -> str:
    """
    Format the date time to the required format for STAC

    :param dt: date time to format
    """
    dt_str = parser.parse(dt).strftime("%Y-%m-%dT%H:%M:%S.%f")

    # convert the string to datetime object
    dt_obj = datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f")

    return dt_obj



"""
Module for managing the Sentinel Hub configuration and data access
"""

import json
from os.path import join, exists
from typing import Optional
from sentinelhub import (SHConfig, 
                         SentinelHubCatalog, 
                         BBox, 
                         bbox_to_dimensions, 
                         CRS, 
                         SentinelHubRequest, 
                         SentinelHubDownloadClient,
                         MimeType)

from ...repos.AuthRepo import AuthRepo
from .parameters import (SHParametersFeature, 
                         sentinel_1_search_parameters, 
                         sentinel_2_l2a_search_parameters)
from .utils import check_time_interval_is_range


SENTINEL_PARAMETERS = {'sentinel-1-grd': sentinel_1_search_parameters,
                       'sentinel-2-l2a': sentinel_2_l2a_search_parameters}

SUPPORTED_SENTINEL_MISSIONS = ('sentinel-2-l1c', 'sentinel-2-l2a', 'sentinel-1-grd')


class SHClient():
    """
    Client class to manage the Sentinel Hub Python interface.
    """

    def __init__(self,
                 sh_client_id: Optional[str] = None, 
                 sh_client_secret: Optional[str] = None
                 ) -> None:
        """
        :param sh_client_id: User's OAuth client ID for Sentinel Hub service.
        :param sh_client_secret: User's OAuth client secret for Sentinel Hub service.
        """
        self.config = SHConfig()
        if sh_client_id and sh_client_secret:
            # If the user has provided the credentials, we should save them
            self.config.sh_client_id = sh_client_id
            self.config.sh_client_secret = sh_client_secret
        else:
            # If the user has not provided the credentials, we should check if
            # the user is logged in
            auth_repo = AuthRepo()
            creds_file = auth_repo.creds_path
            if not exists(creds_file):
                raise ValueError('Your are not logged in and have not provided Sentinel Hub credentials.')
            else:
                creds = json.load(open(creds_file, 'r'))
                if not 'SH_CLIENT_ID' in creds.keys() or not 'SH_CLIENT_SECRET' in creds.keys():
                    raise ValueError('If you already had a Sentinel HUB account before accepting the Terms and Conditions, your SH credentials will NOT appear here. You can retrieve them from you Sentinel HUB dashboard)')
                else:
                    self.config.sh_client_id = creds['SH_CLIENT_ID']
                    self.config.sh_client_secret = creds['SH_CLIENT_SECRET']
        self.catalog = SentinelHubCatalog(config=self.config)

    def search_available_sentinel_data(self,
                                       parameters: SHParametersFeature
                                       ) -> list:
        """
        Search and return the available Sentinel data from a given DataCollection in a selected location and time interval. 

        :param parameters: ParametersFeature object with the needed parameters for the search. The neeeded parameters are:
                    - data_collection: Required.
                    - bbox: Required.
                    - time_interval: Required.
                    - fields: Required.
                    - filter: Optional.
        :return results: list with the obtained results.
        """
        search_iterator = self.catalog.search(
            parameters.data_collection,
            bbox=BBox(parameters.bounding_box, crs=CRS.WGS84),
            time=parameters.time_interval,
            filter=parameters.filter,
            fields=parameters.fields
        )

        results = list(search_iterator)

        return results

    def request_bulk_data(self, 
                    parameters: SHParametersFeature
                    ) -> list:
        """
        Get and return a list with the download information for every location. 
        This is the prefered way to download data, as in order to efficiently 
        download data for all requests in parallel, we extract the download 
        information and we pass it to a download client.

        :param parameters: ParametersFeature object with the needed parameters for the search. The neeeded parameters are:
                    - data_to_download: Dictionary with the information about the data that the user wants to
                        download. If not given, will take the bounding_box and time_interval. It could have two formats:
                            - location_id : list(bounding_box)
                            - location_id : dict(bounding_box: list(bounding_box),
                                                time_interval: list(list(time_interval), list(time_interval) ...)
                                                )
                    - data_collection: Required.
                    - resolution: Required.
                    - data_folder: Required.

        :return process_request: list with the download information for every location
        """
        # Use exactly the needed service url to avoid errors
        self.config.sh_base_url = parameters.data_collection.service_url   
        
        process_requests = list()

        root_folder = parameters.data_folder

        if not parameters.data_to_download:
            bbox = parameters.bounding_box
            time_interval = parameters.time_interval
            if not parameters.bounding_box or not parameters.time_interval:
                raise ValueError('You must provide a bounding box and a time interval')
        else:
            for id, info in parameters.data_to_download.items():
                # We should distinct between the possible input formats
                if isinstance(info, tuple):
                    bbox = BBox(info, crs=CRS.WGS84)
                    time_interval = [None]
                elif isinstance(info, dict):   # Bulk download
                    if not info['bounding_box'] or not info['time_interval']:
                        raise ValueError(f'You must provide a bounding box and a time interval for the location {id}')
                    bbox = BBox(info['bounding_box'], crs=CRS.WGS84)
                    time_interval = info['time_interval']
                    # Check if the time interval is a list of lists
                    # If it is, we will do a request for each time interval
                    # it not, we will do a request for the whole time interval
                    if not check_time_interval_is_range(time_interval):
                        # Search all available data in the time interval
                        search_dict = parameters.data_to_download
                        available_data, non_available_data = self.get_available_data_by_location(search_dict,
                                                                                                parameters.data_collection.api_id)
                        # Add the available data to the data_to_download dictionary
                        parameters.data_to_download[id] = available_data[id]
                        time_interval = parameters.data_to_download[id]['time_interval']
                parameters.bounding_box = bbox

            # Create a different data folder for each request
            _data_folder = join(root_folder, id)

            for time in time_interval:
                # Add the date to the data folder name, if it exists
                data_folder = f'{_data_folder}_{time[1]}' if time else _data_folder
                # Add required parameters to the ParametersFeature, in order to
                # pass it to the request data function
                parameters.data_folder = data_folder
                parameters.time_interval = time
                # Do the request and append it to the requests list 
                request = self.request_data(parameters)
                process_requests.append(request)

        return process_requests
    
    def request_data(self, 
                    parameters: SHParametersFeature
                    ) -> list:
        """
        Request the specified Sentinel Hub data and return the SentinelHubRequest object

        :param parameters: ParametersFeature object with the needed parameters for the search. The neeeded parameters are:
                    - time_interval: Required.
                    - data_collection: Required.
                    - resolution: Required.
                    - data_folder: Required.
                    - bounding_box: Required.
                    - mosaicking_order: Optional.

        """
        return SentinelHubRequest(
                    data_folder=parameters.data_folder,
                    evalscript=parameters.evalscript,
                    input_data=[
                        SentinelHubRequest.input_data(
                            data_collection=parameters.data_collection,
                            time_interval=parameters.time_interval,
                            mosaicking_order=parameters.mosaicking_order
                        )
                    ],
                    responses=[
                                SentinelHubRequest.output_response("default", MimeType.TIFF)
                    ],
                    bbox=parameters.bounding_box,
                    size=bbox_to_dimensions(parameters.bounding_box, parameters.resolution),
                    config=self.config,
                )

    def download_data(self,
                      requests: list
                      ) -> list:
        """
        Download data from Sentinel Hub Services using a list of requests. 
        This is the prefered way to download data, in order to efficiently 
        download data for all requests in parallel.
        
        :param requests: list with SentinelHubRequest objects representing each 
                        request to the Sentinel Hub Services.
        :return: list with the downloaded data
        """
        download_client = SentinelHubDownloadClient(config=self.config)
        download_requests = [request.download_list[0] for request in requests]
        # Download data with multiple threads
        data = download_client.download(download_requests)

        return data

    def get_available_data_by_location(self,
                                       search_data: dict,
                                       sentinel_mission: str
                                       ) -> list:
        """
        Search and return a dict with the available Sentinel data for a dict with given locations and a time intervals.

        :param search_data: dictionary with the data required to search the available imagery in a given location
                and time interval. It must have the following format:
                    {<location_id>: {'bounding_box': list(), 'time_interval': list()}, ... }
        :param sentinel_mission: id of the required Sentinel mission. The value must be <sentinel-1> or <sentinel-2>
        
        :return: available_data: available data for downloading for a given location and time interval
        :return: not_available_data: list with the locations that does not have any available data for the
                given location and time interval
        """
        if sentinel_mission not in SUPPORTED_SENTINEL_MISSIONS:
            raise ValueError(f'The specified Sentinel mission is not valid. The values must be between {SUPPORTED_SENTINEL_MISSIONS}')
        
        parameters = SENTINEL_PARAMETERS[sentinel_mission]

        available_data, not_available_data = dict(), list()
        for location_id, location_info in search_data.items():
            parameters.bounding_box = location_info['bounding_box']
            parameters.time_interval = location_info['time_interval']
            results = self.search_available_sentinel_data(parameters)
            if results:
                # The returning results are composed by a list with format 
                # 'id': <image ID>, properties : {'datetime': <image date>}
                # As we can't make a bulk request with the ID but with the date time,
                # and we need all the available images in a time lapse and not
                # a mosaic, we are going to generate a dict with format
                # 'location_id': <location ID>,
                # {'bounding_box': <image bbox>, 'time_interval': <image date>}
                # This dictionary is digerible by the SHClient
                time_intervals = list()
                for result in results:
                    datetime = result['properties']['datetime'][0:10]
                    time_interval = (datetime, datetime)
                    time_intervals.append(time_interval) if time_interval not in time_intervals else time_intervals
                available_data[location_id] = {'bounding_box': location_info['bounding_box'], 'time_interval': time_intervals}
            else:
                # We should have a trace with the locations without
                # available data
                not_available_data.append(location_id)

        return available_data, not_available_data



"""
Utils for Sentinel Hub access
"""

from typing import List


def check_time_interval_is_range(time_interval: List) -> bool:
    """
    """
    if len(time_interval) == 2:   # Suppose it is a simple list, (from-date, to-date)
        if isinstance(time_interval[0], str) and isinstance(time_interval[1], str):
            if time_interval[0] != time_interval[1]:
                return False
    
    return True



import typer
from .commands import auth, datasets
from . import __version__

app = typer.Typer()

app.add_typer(auth.app, name="auth")
app.add_typer(datasets.app, name="datasets")


@app.command()
def version():
    typer.echo(f"EOTDL Version: {__version__}")


if __name__ == "__main__":
    app()



'''
Module for the STAC label extension ImageNameLabeler object
'''

import pystac
import pandas as pd
import json

from tqdm import tqdm
from os.path import join, dirname
from typing import List, Optional, Union
from ...extent import get_unknow_extent
from .base import LabelExtensionObject
from pystac.extensions.label import LabelExtension


class ImageNameLabeler(LabelExtensionObject):
    def __init__(self) -> None:
        super().__init__()

    def generate_stac_labels(
        self,
        catalog: Union[pystac.Catalog, str],
        stac_dataframe: Optional[pd.DataFrame] = None,
        collection: Optional[Union[pystac.Collection, str]] = 'source',
        label_description: Optional[str] = "Item label",
        label_type: Optional[str] = "vector",
        label_names: Optional[List[str]] = ["label"],
        **kwargs
    ) -> None:
        """
        Generate a labels collection from a STAC dataframe. 
        This class uses the label column of the dataframe as the label names.
        
        :param catalog: catalog to add the labels collection to
        :param stac_dataframe: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param collection: collection to add the labels collection to
        :param label_description: label description
        :param label_type: label type
        :param label_names: list of label names
        :param kwargs: optional arguments
            :param kwargs.label_properties: list of label properties
            :param kwargs.label_methods: list of label methods
            :param kwargs.label_tasks: list of label tasks
        """
        if stac_dataframe.empty:
            raise ValueError(
                "No STAC dataframe provided, please provide a STAC dataframe or generate it with <get_stac_dataframe> method"
            )
        if isinstance(catalog, str):
            catalog = pystac.Catalog.from_file(catalog)

        # Add the labels collection to the catalog
        # If exists a source collection, get it extent
        source_collection = catalog.get_child(collection)
        if source_collection:
            extent = source_collection.extent
            source_items = source_collection.get_stac_objects(pystac.RelType.ITEM)
        else:
            if not collection:
                raise ValueError(
                    "No source collection provided, please provide a source collection"
                )
            extent = get_unknow_extent()

        # Create the labels collection and add it to the catalog if it does not exist
        # If it exists, remove it
        collection = pystac.Collection(id="labels", description="Labels", extent=extent)
        if collection.id in [c.id for c in catalog.get_children()]:
            catalog.remove_child(collection.id)
        catalog.add_child(collection)

        # Generate the labels items
        print("Generating labels collection...")
        for source_item in tqdm(source_items):
            # There must be an item ID column in the STAC dataframe
            if not 'id' in stac_dataframe.columns:
                raise ValueError(
                    "No item ID column found in the STAC dataframe, please provide a STAC dataframe with the item ID column"
                )
            label_classes = stac_dataframe.label.unique().tolist()

            # Create the label item
            label_item = self.add_extension_to_item(
                source_item,
                label_description=label_description,
                label_type=label_type,
                label_names=[label_names],
                label_classes=[label_classes],
                **kwargs
            )
            # Add the self href to the label item, following the Best Practices Layout
            # https://github.com/radiantearth/stac-spec/blob/master/best-practices.md
            label_item.set_self_href(
                join(
                    dirname(collection.get_self_href()),
                    label_item.id,
                    f"{label_item.id}.json"
                    )
            )
            collection.add_item(label_item)

        # Add the extension to the collection
        self.add_extension_to_collection(
            collection,
            label_names=[label_names],
            label_classes=[label_classes],
            label_type=label_type,
        )

        # Validate and save the catalog
        # Before adding the geojson, we need to save the catalog
        # and then iterate over the items to add the geojson
        try:
            pystac.validation.validate(catalog)
            catalog.normalize_and_save(dirname(catalog.get_self_href()), pystac.CatalogType.SELF_CONTAINED)
        except pystac.STACValidationError as e:
            raise pystac.STACError(f"Catalog validation error: {e}")
        
        # Add a GeoJSON FeatureCollection to every label item, as recommended by the spec
        # https://github.com/stac-extensions/label#assets
        self.add_geojson_to_items(collection, 
                                  stac_dataframe,
                                  label_type=label_type)
        catalog.normalize_and_save(dirname(catalog.get_self_href()), pystac.CatalogType.SELF_CONTAINED)
        print('Success on labels generation!')

    def add_geojson_to_items(self, 
                             collection: pystac.Collection,
                             df: pd.DataFrame,
                             label_type: str
                             ) -> None:
        """
        Add a GeoJSON FeatureCollection to every label item, as recommended by the spec
        https://github.com/stac-extensions/label#assets

        :param collection: collection to add the labels collection to
        :param df: dataframe with the STAC metadata of a given directory containing the assets to generate metadata
        :param label_type: label type
        """
        for item in collection.get_all_items():
            geojson_path = join(dirname(item.get_self_href()), f'{item.id}.geojson')

            properties = {'roles': ['labels', f'labels-{label_type}']}

            # TODO depending on the tasks, there must be extra fields
            # https://github.com/stac-extensions/label#assets
            tasks = item.properties['label:tasks']
            if 'tile_regression' in tasks:
                pass
            elif any(task in tasks for task in ('tile_classification', 'object_detection', 'segmentation')):
                pass

            label_ext = LabelExtension.ext(item)
            label_ext.add_geojson_labels(href=geojson_path, 
                                         title='Label', 
                                         properties=properties)
            item.make_asset_hrefs_relative()
            
            item_id = item.id
            geometry = item.geometry
            labels = [df[df['id'] == item_id]['label'].values[0]]
            # There is data like DEM data that does not have datetime but start and end datetime
            datetime = item.datetime.isoformat() if item.datetime else (item.properties.start_datetime.isoformat(),
                                                                        item.properties.end_datetime.isoformat())
            labels_properties = dict(zip(item.properties['label:properties'], labels)) if label_type == 'vector' else dict()
            labels_properties['datetime'] = datetime

            geojson = {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "geometry": geometry,
                        "properties": labels_properties,
                    }
                ],
            }

            with open(geojson_path, "w") as f:
                json.dump(geojson, f)



import requests
import os
from tqdm import tqdm

from ..repos import APIRepo


class FilesAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def ingest_file(self, file, dataset_id, version, parent, id_token, checksum=None):
        reponse = requests.post(
            self.url + "datasets/" + dataset_id,
            files={"file": open(file, "rb")},
            data={"checksum": checksum, "version": version, "parent": parent}
            if checksum
            else None,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(reponse)

    def ingest_existing_file(
        self,
        filename,
        dataset_id,
        version,
        file_version,
        id_token,
        checksum=None,
    ):
        reponse = requests.post(
            self.url + "datasets/" + dataset_id,
            data={
                "checksum": checksum,
                "version": version,
                "filename": filename,
                "fileversion": file_version,
            }
            if checksum
            else None,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(reponse)

    def retrieve_dataset_files(self, dataset_id, version=None):
        url = self.url + "datasets/" + dataset_id + "/files"
        if version is not None:
            url += "?version=" + str(version)
        response = requests.get(url)
        return self.format_response(response)

    def download_file(self, dataset_id, file_name, id_token, path, file_version):
        url = self.url + "datasets/" + dataset_id + "/download/" + file_name
        if file_version is not None:
            url += "?version=" + str(file_version)
        return self.download_file_url(url, file_name, path, id_token)

    def download_file_url(self, url, filename, path, id_token, progress=False):
        headers = {"Authorization": "Bearer " + id_token}
        path = f"{path}/{filename}"
        for i in range(1, len(path.split("/")) - 1):
            # print("/".join(path.split("/")[: i + 1]))
            os.makedirs("/".join(path.split("/")[: i + 1]), exist_ok=True)
        with requests.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024 * 1024 * 10
            if progress:
                progress_bar = tqdm(
                    total=total_size, unit="iB", unit_scale=True, unit_divisor=1024
                )
            with open(path, "wb") as f:
                for chunk in r.iter_content(block_size):
                    if progress:
                        progress_bar.update(len(chunk))
                    if chunk:
                        f.write(chunk)
            if progress:
                progress_bar.close()
            return path

    # def ingest_file_url(self, file, dataset, id_token):
    #     reponse = requests.post(
    #         self.url + f"datasets/{dataset}/url",
    #         json={"url": file},
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if reponse.status_code != 200:
    #         return None, reponse.json()["detail"]
    #     return reponse.json(), None

    # def read_in_chunks(self, file_object, CHUNK_SIZE):
    #     while True:
    #         data = file_object.read(CHUNK_SIZE)
    #         if not data:
    #             break
    #         yield data

    # def prepare_large_upload(self, file, dataset_id, checksum, id_token):
    #     filename = Path(file).name
    #     response = requests.post(
    #         self.url + f"datasets/{dataset_id}/uploadId",
    #         json={"name": filename, "checksum": checksum},
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if response.status_code != 200:
    #         raise Exception(response.json()["detail"])
    #     data = response.json()
    #     upload_id, parts = (
    #         data["upload_id"],
    #         data["parts"] if "parts" in data else [],
    #     )
    #     return upload_id, parts

    # def get_chunk_size(self, content_size):
    #     # adapt chunk size to content size to avoid S3 limits (10000 parts, 500MB per part, 5TB per object)
    #     chunk_size = 1024 * 1024 * 10  # 10 MB (up to 100 GB, 10000 parts)
    #     if content_size >= 1024 * 1024 * 1024 * 100:  # 100 GB
    #         chunk_size = 1024 * 1024 * 100  # 100 MB (up to 1 TB, 10000 parts)
    #     elif content_size >= 1024 * 1024 * 1024 * 1000:  # 1 TB
    #         chunk_size = 1024 * 1024 * 500  # 0.5 GB (up to 5 TB, 10000 parts)
    #     return chunk_size

    # def ingest_large_dataset(self, file, upload_id, id_token, parts):
    #     content_path = os.path.abspath(file)
    #     content_size = os.stat(content_path).st_size
    #     chunk_size = self.get_chunk_size(content_size)
    #     total_chunks = content_size // chunk_size
    #     # upload chunks sequentially
    #     pbar = tqdm(
    #         self.read_in_chunks(open(content_path, "rb"), chunk_size),
    #         total=total_chunks,
    #     )
    #     index = 0
    #     for chunk in pbar:
    #         part = index // chunk_size + 1
    #         offset = index + len(chunk)
    #         index = offset
    #         if part not in parts:
    #             checksum = hashlib.md5(chunk).hexdigest()
    #             response = requests.post(
    #                 self.url + "datasets/chunk/" + upload_id,
    #                 files={"file": chunk},
    #                 data={"part_number": part, "checksum": checksum},
    #                 headers={"Authorization": "Bearer " + id_token},
    #             )
    #             if response.status_code != 200:
    #                 raise Exception(response.json()["detail"])
    #         pbar.set_description(
    #             "{:.2f}/{:.2f} MB".format(
    #                 offset / 1024 / 1024, content_size / 1024 / 1024
    #             )
    #         )
    #     pbar.close()
    #     return

    # def complete_upload(self, id_token, upload_id):
    #     r = requests.post(
    #         self.url + "datasets/complete/" + upload_id,
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if r.status_code != 200:
    #         return None, r.json()["detail"]
    #     return r.json(), None

    # def update_dataset(self, name, path, id_token, checksum):
    #     # check that dataset exists
    #     data, error = self.retrieve_dataset(name)
    #     if error:
    #         return None, error
    #     # first call to get upload id
    #     dataset_id = data["id"]
    #     url = self.url + f"datasets/chunk/{dataset_id}?checksum={checksum}"
    #     response = requests.get(url, headers={"Authorization": "Bearer " + id_token})
    #     if response.status_code != 200:
    #         return None, response.json()["detail"]
    #     data = response.json()
    #     _, upload_id, parts = data["dataset_id"], data["upload_id"], data["parts"]
    #     # assert dataset_id is None
    #     content_path = os.path.abspath(path)
    #     content_size = os.stat(content_path).st_size
    #     url = self.url + "datasets/chunk"
    #     chunk_size = 1024 * 1024 * 100  # 100 MiB
    #     total_chunks = content_size // chunk_size
    #     headers = {
    #         "Authorization": "Bearer " + id_token,
    #         "Upload-Id": upload_id,
    #         "Dataset-Id": dataset_id,
    #     }
    #     # upload chunks sequentially
    #     pbar = tqdm(
    #         self.read_in_chunks(open(content_path, "rb"), chunk_size),
    #         total=total_chunks,
    #     )
    #     index = 0
    #     for chunk in pbar:
    #         offset = index + len(chunk)
    #         part = index // chunk_size + 1
    #         index = offset
    #         if part not in parts:
    #             headers["Part-Number"] = str(part)
    #             file = {"file": chunk}
    #             r = requests.post(url, files=file, headers=headers)
    #             if r.status_code != 200:
    #                 return None, r.json()["detail"]
    #         pbar.set_description(
    #             "{:.2f}/{:.2f} MB".format(
    #                 offset / 1024 / 1024, content_size / 1024 / 1024
    #             )
    #         )
    #     pbar.close()
    #     # complete upload
    #     url = self.url + "datasets/complete"
    #     r = requests.post(
    #         url,
    #         json={"checksum": checksum},
    #         headers={
    #             "Authorization": "Bearer " + id_token,
    #             "Upload-Id": upload_id,
    #             "Dataset-Id": dataset_id,
    #         },
    #     )
    #     if r.status_code != 200:
    #         return None, r.json()["detail"]
    #     return r.json(), None

    # def delete_file(self, dataset_id, file_name, id_token):
    #     response = requests.delete(
    #         self.url + "datasets/" + dataset_id + "/file/" + file_name,
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if response.status_code != 200:
    #         return None, response.json()["detail"]
    #     return response.json(), None

    # def ingest_stac(self, stac_json, dataset_id, id_token):
    #     reponse = requests.put(
    #         self.url + f"datasets/stac/{dataset_id}",
    #         json={"stac": stac_json},
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if reponse.status_code != 200:
    #         return None, reponse.json()["detail"]
    #     return reponse.json(), None

    # def download_stac(self, dataset_id, id_token):
    #     url = self.url + "datasets/" + dataset_id + "/download"
    #     headers = {"Authorization": "Bearer " + id_token}
    #     response = requests.get(url, headers=headers)
    #     if response.status_code != 200:
    #         return None, response.json()["detail"]
    #     return gpd.GeoDataFrame.from_features(response.json()["features"]), None



"""
Module for EO-TDL metadata
"""

from typing import Optional

from os import remove
from glob import glob
from os.path import join


def remove_raster_metadata(folder: str, metadata_file: Optional[str] = 'metadata.json') -> None:
    """
    Remove metadata.json file from a folder

    :param folder: folder path
    :param metadata_file: metadata file name
    """
    # Search for all the metadata files in the folder
    metadata_files = glob(join(folder, "**", metadata_file), recursive=True)
    # Remove all the metadata files
    for metadata_file in metadata_files:
        remove(metadata_file)



"""Implements the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`."""

import traceback
import json
import random

import pystac

from tqdm import tqdm
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.label import LabelExtension
from pystac import STACValidationError
from shutil import rmtree
from os.path import dirname, exists
from pystac.cache import ResolvedObjectCache
from pystac.extensions.hooks import ExtensionHooks
from typing import Any, Dict, List, Optional, Generic, TypeVar, Union, Set
from ..utils import make_links_relative_to_path

T = TypeVar("T", pystac.Item, pystac.Collection, pystac.Catalog)


SCHEMA_URI: str = "https://raw.githubusercontent.com/earthpulse/ml-dataset/main/json-schema/schema.json"
PREFIX: str = "ml-dataset:"


class MLDatasetExtension(
    pystac.Catalog,
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[
        Union[pystac.item.Item, pystac.collection.Collection, pystac.catalog.Catalog]
    ],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection`, :class:`~pystac.Item`, or :class:`~pystac.Catalog` with
    properties from the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`MLDatasetExtension`, use the
    :meth:`MLDatasetExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> ml_ext = MLDatasetExtension.ext(item)
    """

    catalog: pystac.Catalog
    """The :class:`~pystac.Catalog` being extended."""

    properties: Dict[str, Any]
    """The :class:`~pystac.Catalog` extra fields, including extension properties."""

    links: List[pystac.Link]
    """The list of :class:`~pystac.Link` objects associated with the
    :class:`~pystac.Catalog` being extended, including links added by this extension.
    """

    def __init__(self, catalog: pystac.Catalog):
        super().__init__(id=catalog.id, description=catalog.description)
        self._catalog = catalog
        self.id = catalog.id
        self.description = catalog.description
        self.title = catalog.title if catalog.title else None
        self.stac_extensions = (
            catalog.stac_extensions if catalog.stac_extensions else []
        )
        self.extra_fields = self.properties = (
            catalog.extra_fields if catalog.extra_fields else {}
        )
        self.links = catalog.links
        self._resolved_objects = ResolvedObjectCache()

    def apply(self, name: str = None) -> None:
        self.name = name

    @property
    def name(self) -> str:
        return self.extra_fields[f"{PREFIX}name"]

    @name.setter
    def name(self, v: str) -> None:
        self.extra_fields[f"{PREFIX}name"] = v

    @property
    def tasks(self) -> List:
        return self.extra_fields[f"{PREFIX}tasks"]

    @tasks.setter
    def tasks(self, v: Union[list, tuple]) -> None:
        self.extra_fields[f"{PREFIX}tasks"] = v

    @property
    def type(self) -> str:
        return self.extra_fields[f"{PREFIX}type"]

    @type.setter
    def type(self, v: str) -> None:
        self.extra_fields[f"{PREFIX}type"] = v

    @property
    def inputs_type(self) -> str:
        return self.extra_fields[f"{PREFIX}inputs-type"]

    @inputs_type.setter
    def inputs_type(self, v: str) -> None:
        self.extra_fields[f"{PREFIX}inputs-type"] = v

    @property
    def annotations_type(self) -> str:
        return self.extra_fields[f"{PREFIX}annotations-type"]

    @annotations_type.setter
    def annotations_type(self, v: str) -> None:
        self.extra_fields[f"{PREFIX}annotations-type"] = v

    @property
    def splits(self) -> List[str]:
        self.extra_fields[f"{PREFIX}splits"]

    @splits.setter
    def splits(self, v: dict) -> None:
        self.extra_fields[f"{PREFIX}splits"] = v

    @property
    def quality_metrics(self) -> List[dict]:
        self.extra_fields[f"{PREFIX}quality-metrics"]

    @quality_metrics.setter
    def quality_metrics(self, v: dict) -> None:
        self.extra_fields[f"{PREFIX}quality-metrics"] = v

    @property
    def version(self) -> str:
        self.extra_fields[f"{PREFIX}version"]

    @version.setter
    def version(self, v: str) -> None:
        self.extra_fields[f"{PREFIX}version"] = v

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    def add_metric(self, metric: dict) -> None:
        """Add a metric to this object's set of metrics.

        Args:
             metric : The metric to add.
        """
        if not self.extra_fields.get(f'{PREFIX}quality-metrics'):
            self.extra_fields[f'{PREFIX}quality-metrics'] = []

        if metric not in self.extra_fields[f'{PREFIX}quality-metrics']:
            self.extra_fields[f'{PREFIX}quality-metrics'].append(metric)

    def add_metrics(self, metrics: List[dict]) -> None:
        """Add a list of metrics to this object's set of metrics.

        Args:
             metrics : The metrics to add.
        """
        for metric in metrics:
            self.add_metric(metric)

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False):
        """Extends the given STAC Object with properties from the
        :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.

        This extension can be applied to instances of :class:`~pystac.Catalog`,
        :class:`~pystac.Collection` or :class:`~pystac.Item`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.validate_has_extension(obj, add_if_missing)
            return CollectionMLDatasetExtension(obj)
        elif isinstance(obj, pystac.Catalog):
            cls.validate_has_extension(obj, add_if_missing)
            return MLDatasetExtension(obj)
        elif isinstance(obj, pystac.Item):
            cls.validate_has_extension(obj, add_if_missing)
            return ItemMLDatasetExtension(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class CollectionMLDatasetExtension(MLDatasetExtension[pystac.Collection]):
    """A concrete implementation of :class:`MLDatasetExtension` on an
    :class:`~pystac.Collection` that extends the properties of the Collection to include
    properties defined in the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.
    """

    collection: pystac.Collection
    properties: Dict[str, Any]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields
        self.properties[f"{PREFIX}split-items"] = []

    def __repr__(self) -> str:
        return "<CollectionMLDatasetExtension Item id={}>".format(self.collection.id)

    @property
    def splits(self) -> List[dict]:
        return self._splits

    @splits.setter
    def splits(self, v: dict) -> None:
        self.properties[f"{PREFIX}split-items"] = v

    def add_split(self, v: dict) -> None:
        self.properties[f"{PREFIX}split-items"].append(v)

    def create_and_add_split(
        self, split_data: List[pystac.Item], split_type: str
    ) -> None:
        """ """
        items_ids = [item.id for item in split_data]
        items_ids.sort()

        split = {"name": split_type, "items": items_ids}

        if not self.properties.get(f"{PREFIX}split-items"):
            self.properties[f"{PREFIX}split-items"] = []

        self.add_split(split)
        print(f"Generating {split_type} split...")
        for _item in tqdm(split_data):
            item = self.collection.get_item(_item.id)
            if item:
                item_ml = MLDatasetExtension.ext(item, add_if_missing=True)
                item_ml.split = split_type



class ItemMLDatasetExtension(MLDatasetExtension[pystac.Item]):
    """A concrete implementation of :class:`MLDatasetExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`MLDatasetExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: Dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    @property
    def split(self) -> str:
        return self._split

    @split.setter
    def split(self, v: str) -> None:
        self.properties[f"{PREFIX}split"] = v

    def __repr__(self) -> str:
        return "<ItemMLDatasetExtension Item id={}>".format(self.item.id)


class MLDatasetQualityMetrics:
    """ """

    @classmethod
    def calculate(self, catalog: Union[pystac.Catalog, str]) -> None:
        """ """

        if isinstance(catalog, str):
            catalog = MLDatasetExtension(pystac.read_file(catalog))
        # Check the catalog has the extension
        if not MLDatasetExtension.has_extension(catalog):
            raise pystac.ExtensionNotImplemented(
                f"MLDatasetExtension does not apply to type '{type(catalog).__name__}'"
            )

        catalog.make_all_asset_hrefs_absolute()
        try:
            catalog.add_metric(self._search_spatial_duplicates(catalog))
            catalog.add_metric(self._get_classes_balance(catalog))
        except AttributeError:
            raise pystac.ExtensionNotImplemented(
                f"The catalog does not have the required properties or the ML-Dataset extension to calculate the metrics"
            )
        finally:
            catalog.make_all_asset_hrefs_relative()
            
        try:
            print("Validating and saving...")
            catalog.validate()
            destination = dirname(catalog.get_self_href())
            rmtree(
                destination
            )  # Remove the old catalog and replace it with the new one
            catalog.save(dest_href=destination)
            print("Success!")
        except STACValidationError as error:
            # Return full callback
            traceback.print_exc()

    @staticmethod
    def _search_spatial_duplicates(catalog: pystac.Catalog):
        """ """
        print("Looking for spatial duplicates...")
        items = list(
            set(
                [item 
                 for item in tqdm(catalog.get_items(recursive=True)) 
                 if not LabelExtension.has_extension(item)
                 ]
                )
            )

        # Initialize the spatial duplicates dict
        spatial_duplicates = {"name": "spatial-duplicates", "values": [], "total": 0}

        items_bboxes = dict()
        for item in items:
            # Get the item bounding box
            bbox = str(item.bbox)
            # If the bounding box is not in the items dict, add it
            if bbox not in items_bboxes.keys():
                items_bboxes[bbox] = item.id
            # If the bounding box is already in the items dict, add it to the duplicates dict
            else:
                spatial_duplicates["values"].append(
                    {"item": item.id, "duplicate": items_bboxes[bbox]}
                )
                spatial_duplicates["total"] += 1

        return spatial_duplicates

    @staticmethod
    def _get_classes_balance(catalog: pystac.Catalog) -> dict:
        """ """

        def get_label_properties(items: List[pystac.Item]) -> List:
            """
            """
            label_properties = list()
            for label in items:
                label_ext = LabelExtension.ext(label)
                for prop in label_ext.label_properties:
                    if prop not in label_properties:
                        label_properties.append(prop)

            return label_properties
    
        catalog.make_all_asset_hrefs_absolute()

        labels = list(
            set(
                [item 
                 for item in tqdm(catalog.get_items(recursive=True), desc="Calculating classes balance...") 
                 if LabelExtension.has_extension(item)
                 ]
                )
            )

        # Initialize the classes balance dict
        classes_balance = {"name": "classes-balance", "values": []}
        label_properties = get_label_properties(labels)

        for property in label_properties:
            property_balance = {"name": property, "values": []}
            properties = dict()
            for label in labels:
                asset_path = label.assets["labels"].href
                print(asset_path)
                # Open the linked geoJSON to obtain the label properties
                try:
                    with open(asset_path) as f:
                        label_data = json.load(f)
                except FileNotFoundError:
                    raise FileNotFoundError(
                        f"The file {asset_path} does not exist. Make sure the assets hrefs are correct"
                    )
                # Get the property
                for feature in label_data["features"]:
                    if property in feature["properties"]:
                        property_value = feature["properties"][property]
                    else:
                        if feature["properties"]["labels"]:
                            property_value = feature["properties"]["labels"][0]
                        else:
                            continue
                    if property_value not in properties:
                        properties[property_value] = 0
                    properties[property_value] += 1
            
            # Create the property balance dict
            total_labels = sum(properties.values())
            for key, value in properties.items():
                property_balance["values"].append(
                    {
                        "class": key,
                        "total": value,
                        "percentage": int(value / total_labels * 100),
                    }
                )

            classes_balance["values"].append(property_balance)

        catalog.make_all_asset_hrefs_relative()

        return classes_balance


class MLDatasetExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: Set[str] = set()
    stac_object_types = {
        pystac.STACObjectType.CATALOG,
        pystac.STACObjectType.COLLECTION,
        pystac.STACObjectType.ITEM,
    }


STORAGE_EXTENSION_HOOKS: ExtensionHooks = MLDatasetExtensionHooks()


def add_ml_extension(
    catalog: Union[pystac.Catalog, str],
    destination: Optional[str] = None,
    splits: Optional[bool] = False,
    splits_collection_id: Optional[str] = "labels",
    splits_names: Optional[list] = ("Training", "Validation", "Test"),
    split_proportions: Optional[List[int]] = (80, 10, 10),
    catalog_type: Optional[pystac.CatalogType] = pystac.CatalogType.SELF_CONTAINED,
    **kwargs,
) -> None:
    """
    Adds the ML Dataset extension to a STAC catalog.

    Args:
        catalog : The STAC catalog to add the extension to.
        destination : The destination path to save the catalog to.
        splits : The splits to make.
        split_proportions : The proportions of the splits.
    """
    if not isinstance(catalog, pystac.Catalog) and isinstance(catalog, str):
        catalog = pystac.read_file(catalog)
    elif isinstance(catalog, pystac.Catalog):
        pass
    else:
        raise pystac.ExtensionTypeError(
            f"MLDatasetExtension does not apply to type '{type(catalog).__name__}'"
        )

    catalog_ml_dataset = MLDatasetExtension.ext(catalog, add_if_missing=True)
    catalog_ml_dataset.set_self_href(destination + "/catalog.json")
    catalog_ml_dataset.set_root(catalog_ml_dataset)

    # Set extension properties
    for key, value in kwargs.items():
        setattr(catalog_ml_dataset, key, value)

    # Make splits if needed
    if splits:
        catalog_ml_dataset.splits = splits_names  # Add the splits names to the catalog
        train_size, test_size, val_size = split_proportions
        splits_collection = catalog.get_child(
            splits_collection_id
        )  # Get the collection to split
        make_splits(
            splits_collection,
            train_size=train_size,
            test_size=test_size,
            val_size=val_size,
            **kwargs,
        )

    # Normalize the ref on the same folder
    if destination:
        make_links_relative_to_path(destination, catalog_ml_dataset)
        # TODO not working
    else:
        destination = dirname(catalog.get_self_href())

    try:
        print("Validating and saving...")
        catalog_ml_dataset.validate()
        rmtree(destination) if exists(destination) else None # Remove the old catalog and replace it with the new one
        catalog_ml_dataset.normalize_and_save(root_href=destination, 
                                              catalog_type=catalog_type
                                              )
        print("Success!")
    except STACValidationError:
        # Return full callback
        traceback.print_exc()


def make_splits(
    labels_collection: Union[CollectionMLDatasetExtension, pystac.Collection, str],
    splits_names: Optional[List[str]] = ("Training", "Validation", "Test"),
    splits_proportions: Optional[List[int]] = (80, 10, 10),
    verbose: Optional[bool] = True,
    **kwargs,
) -> None:
    """
    Makes the splits of the labels collection.

    Args:
        labels_collection : The STAC Collection make the splits on.
        train_size : The percentage of the dataset to use for training.
        test_size : The percentage of the dataset to use for testing.
        val_size : The percentage of the dataset to use for validation.
        verbose : Whether to print the sizes of the splits.
    """
    if isinstance(labels_collection, str):
        labels_collection = pystac.read_file(labels_collection)

    train_size, test_size, val_size = splits_proportions

    if train_size + test_size + val_size != 100:
        raise ValueError("The sum of the splits must be 100")

    # Get all items in the labels collection
    items = [item for item in labels_collection.get_items(recursive=True)]

    # Calculate indices to split the items
    length = len(items)
    idx_train = int(train_size / 100 * length)
    idx_test = int(test_size / 100 * length)
    if val_size:
        idx_val = int(val_size / 100 * length)

    print("Generating splits...")
    if verbose:
        print(f"Total size: {length}")
        print(f"Train size: {idx_train}")
        print(f"Test size: {idx_test}")
        if val_size:
            print(f"Validation size: {idx_val}")

    # Make sure the items are shuffled
    random.shuffle(items)

    # Split the items
    train_items = items[:idx_train]
    test_items = items[idx_train : idx_train + idx_test]
    if val_size:
        val_items = items[idx_train + idx_test : idx_train + idx_test + idx_val]

    # Create the splits in the collection
    labels_collection = MLDatasetExtension.ext(labels_collection, add_if_missing=True)
    for split_type, split_data in zip(
        splits_names, [train_items, test_items, val_items]
    ):
        labels_collection.create_and_add_split(split_data, split_type)

    print("Success on splits generation!")



from .sar import SarExtensionObject
from .raster import RasterExtensionObject
from .projection import ProjExtensionObject
from .dem import DEMExtensionObject
from .eo import EOS2ExtensionObject
from .label import LabelExtensionObject, ImageNameLabeler, ScaneoLabeler
from .ml_dataset import add_ml_extension, MLDatasetQualityMetrics


SUPPORTED_EXTENSIONS = ('eo', 'sar', 'proj', 'raster')


type_stac_extensions_dict = {
    "sar": SarExtensionObject(),
    "eo": EOS2ExtensionObject(),
    "dem": DEMExtensionObject(),
    "raster": RasterExtensionObject(),
    "proj": ProjExtensionObject()
}



'''
Geometry utils
'''

from pandas import isna


def convert_df_geom_to_shape(row):
    """
    Convert the geometry of a dataframe row to a shapely shape

    :param row: row of a dataframe
    """
    from shapely.geometry import shape

    if not isna(row["geometry"]):
        geo = shape(row["geometry"])
        wkt = geo.wkt
    else:
        wkt = "POLYGON EMPTY"

    return wkt


'''
Paths utils
'''

from os.path import dirname
from typing import Union, Optional
from glob import glob


def count_ocurrences(text: str, text_list: list) -> int:
    """
    Count the number of ocurrences of a string in a list of strings

    :param text: string to count the ocurrences
    """
    count = 0
    for string in text_list:
        if text in string:
            count += 1
    return count


def cut_images(images_list: Union[list, tuple]) -> list:
    """
    Get the directory name of each image and return a list of unique directories

    :param images_list: list of images

    :return: list of unique directories
    """
    dirnames = list()
    images = list()

    for image in images_list:
        dir = dirname(image)
        if dir not in dirnames:
            dirnames.append(dir)
            images.append(image)

    return images


def get_all_images_in_path(path: str, image_format: Optional[str] = 'tif') -> list:
        """
        Get all the images in a directory

        :param path: path to the directory

        :return: list of images
        """
        return glob(str(path) + f'/**/*.{image_format}', recursive=True)



from pydantic import BaseModel, validator
from typing import List


class Metadata(BaseModel):
    authors: List[str]
    license: str
    source: str
    name: str

    # validate source is a URL
    @validator("source")
    def source_is_url(cls, v):
        if not v.startswith("http") and not v.startswith("https"):
            raise ValueError("source must be a URL")
        return v



'''
Metadata utilities for STAC
'''

import json

from os.path import dirname, join, exists, basename
from os import listdir


def get_item_metadata(raster_path: str) -> str:
    """
    Get the metadata JSON file of a given directory, associated to a raster file

    :param raster_path: path to the raster file
    """
    # Get the directory of the raster file
    raster_dir_path = dirname(raster_path)
    # Get the metadata JSON file
    # Check if there is a metadata.json file in the directory
    if 'metadata.json' in listdir(raster_dir_path):
        metadata_json = join(raster_dir_path, 'metadata.json')
    else:
        # If there is no metadata.json file in the directory, check if there is
        # a json file with the same name as the raster file
        raster_name = basename(raster_path).split('.')[0]
        metadata_json = join(raster_dir_path, f'{raster_name}.json')
        if not exists(metadata_json):
            # If there is no metadata file in the directory, return None
            return None
    
    # Open the metadata file and return it
    with open(metadata_json, 'r') as f:
        metadata = json.load(f)
    
    return metadata



from .sentinelhub import (SHClient, 
                          SHParametersFeature, 
                          EvalScripts, 
                          sentinel_1_download_parameters, 
                          sentinel_2_l2a_download_parameters,
                          dem_download_parameters)
from .airbus import (AirbusClient,
                     get_airbus_access_token)


'''
Module for STAC extent
'''

import pystac
from datetime import datetime
import rasterio
import json

from glob import glob
from os.path import dirname
from typing import List

from .utils import get_item_metadata


def get_dem_temporal_interval() -> pystac.TemporalExtent:
    """
    Get a temporal interval for DEM data
    """
    min_date = datetime.strptime('2011-01-01', '%Y-%m-%d')
    max_date = datetime.strptime('2015-01-07', '%Y-%m-%d')

    return pystac.TemporalExtent([(min_date, max_date)])
    
def get_unknow_temporal_interval() -> pystac.TemporalExtent:
    """
    Get an unknown temporal interval
    """
    min_date = datetime.strptime('2000-01-01', '%Y-%m-%d')
    max_date = datetime.strptime('2023-12-31', '%Y-%m-%d')

    return pystac.TemporalExtent([(min_date, max_date)])
    
def get_unknow_extent() -> pystac.Extent:
    """
    """
    return pystac.Extent(spatial=pystac.SpatialExtent([[0, 0, 0, 0]]),
                         temporal=pystac.TemporalExtent([(datetime.strptime('2000-01-01', '%Y-%m-%d'), 
                                                          datetime.strptime('2023-12-31', '%Y-%m-%d')
                                                          )]))


def get_collection_extent(rasters: List[str]) -> pystac.Extent:
    """
    Get the extent of a collection
    
    :param rasters: list of rasters
    """
    # Get the spatial extent of the collection
    spatial_extent = get_collection_spatial_extent(rasters)
    # Get the temporal interval of the collection
    temporal_interval = get_collection_temporal_interval(rasters)
    # Create the Extent object
    extent = pystac.Extent(spatial=spatial_extent, temporal=temporal_interval)

    return extent
    
def get_collection_spatial_extent(rasters: List[str]) -> pystac.SpatialExtent:
    """
    Get the spatial extent of a collection

    :param path: path to the directory
    """
    # Get the bounding boxes of all the given rasters
    bboxes = list()
    for raster in rasters:
        with rasterio.open(raster) as ds:
            bounds = ds.bounds
            dst_crs = 'EPSG:4326'
            try:
                left, bottom, right, top = rasterio.warp.transform_bounds(ds.crs, dst_crs, *bounds)
                bbox = [left, bottom, right, top]
            except rasterio.errors.CRSError:
                spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
                return spatial_extent
            bboxes.append(bbox)
    # Get the minimum and maximum values of the bounding boxes
    try:
        left = min([bbox[0] for bbox in bboxes])
        bottom = min([bbox[1] for bbox in bboxes])
        right = max([bbox[2] for bbox in bboxes])
        top = max([bbox[3] for bbox in bboxes])
        spatial_extent = pystac.SpatialExtent([[left, bottom, right, top]])
    except ValueError:
        spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
    finally:
        return spatial_extent

def get_collection_temporal_interval(rasters: List[str]) -> pystac.TemporalExtent:
    """
    Get the temporal interval of a collection

    :param path: path to the directory
    """
    # Get all the metadata.json files in the directory of all the given rasters
    metadata_jsons = list()
    for raster in rasters:
        metadata_json = get_item_metadata(raster)
        if metadata_json:
            metadata_jsons.append(metadata_json)

    if not metadata_jsons:
        return get_unknow_temporal_interval()   # If there is no metadata, set a generic temporal interval
    
    # Get the temporal interval of every metadata.json file and the type of the data
    data_types = list()
    temporal_intervals = list()
    for metadata in metadata_jsons:
        # Append the temporal interval to the list as a datetime object
        temporal_intervals.append(metadata['acquisition-date']) if metadata['acquisition-date'] else None
        # Append the data type to the list
        data_types.append(metadata['type']) if metadata['type'] else None
        
    if temporal_intervals:
        try:
            # Get the minimum and maximum values of the temporal intervals
            min_date = min([datetime.strptime(interval, '%Y-%m-%d') for interval in temporal_intervals])
            max_date = max([datetime.strptime(interval, '%Y-%m-%d') for interval in temporal_intervals])
        except ValueError:
            min_date = datetime.strptime('2000-01-01', '%Y-%m-%d')
            max_date = datetime.strptime('2023-12-31', '%Y-%m-%d')
        finally:
            # Create the temporal interval
            return pystac.TemporalExtent([(min_date, max_date)])
    else:
        # Check if the collection is composed by DEM data. If not, set a generic temporal interval
        if set(data_types) == {'dem'} or set(data_types) == {'DEM'} or set(data_types) == {'dem', 'DEM'}:
            return get_dem_temporal_interval()
        else:
            return get_unknow_temporal_interval()



"""
Module for formatter classes
"""

import json
import rasterio 

from typing import Optional
from os.path import join, dirname
from os import makedirs
from glob import glob
from shutil import rmtree, copyfile
from .base import FolderFormatter


class SHFolderFormatter(FolderFormatter):
    """
    Class for formatting the directories of Sentinel images downloaded from Sentinel Hub Services
    """

    def __init__(self, root_folder):
        super().__init__()
        self.root = root_folder

    @classmethod
    def structured_format_folders(self) -> None:
        """
        Format the folder structure of the downloaded data from Sentinel Hub to a 
        digestible format, suitable for STAC metadata generation.

        This formatted format is a structured format, where every image is located in
        a single folder, with its metadata.
        """
        images = glob(join(self.root, '**/*.tiff'), recursive=True)
        
        for image in images:
            # Folder with the request and response, with the if of the request
            image_dir = dirname(image)
            # Destination folder, with format <id>_<date>
            dest_dir = dirname(image_dir)
            # Get the request.json file with the request parameters
            request_file = join(image_dir, 'request.json')
            request_json = json.load(open(request_file))
            # Generate a metadata.json file of the raster in the destination folder
            # It will extract some needed parameters, which will be used later 
            # for the STAC generation
            # It also returns the data type of the response
            data_type = self.generate_raster_metadata(request_json, image, dest_dir)       
            # Copy the response tiff to the destination folder
            copyfile(image, join(dest_dir, f'{data_type}.tiff'))
            # Remove the request and response folder
            rmtree(image_dir)

    def unestructured_format_folders(self, output_folder: Optional[str] = None) -> None:
        """
         Format the folder structure of the downloaded data from Sentinel Hub to a 
        digestible format, suitable for STAC metadata generation and labeling with SCANEO.
        
        This formatted format is a unestructured format, where all the images are located
        into the same folder, with them metadata files with the same name as the image.
        """
        images = glob(join(self.root, '**/*.tiff'), recursive=True)

        if not output_folder:
            output_folder = self.root
        makedirs(output_folder, exist_ok=True)

        n = 0
        for image in images:
            # Folder with the request and response, with the id of the request
            image_dir = dirname(image)
            # Parent image folder, with format <id>_<date>
            parent_dir = dirname(image_dir)
            parent_dir_name = parent_dir.split('/')[-1]
            parent_dir_name_split = parent_dir_name.split('_')
            id = parent_dir_name_split[0]   # ID of the given location
            # Get the request.json file with the request parameters
            request_file = join(image_dir, 'request.json')
            request_json = json.load(open(request_file))
            # Generate a metadata.json file of the raster in the destination folder
            # It will extract some needed parameters, which will be used later 
            # for the STAC generation
            # As file name, we give the same name as de image, which is <id>_<n>
            n += 1
            filename = f'{id}_{n}'
            self.generate_raster_metadata(request_json, image, output_folder, filename)    
            # Copy the response tiff to the destination folder
            copyfile(image, join(output_folder, f'{filename}.tif'))
            # Remove the request and response folder
            rmtree(parent_dir)

    def generate_raster_metadata(self,
                                 request_json: dict,
                                 raster_path: str,
                                 output_folder: str,
                                 output_name: Optional[str] = 'metadata'
                                 ) -> None:
        """
        Generate metadata.json file for a raster file

        :param raster_path: path to the raster file
        :param output_folder: output folder to write the metadata.json file to
        :param date_adquired: date adquired of the raster file
        :param output_name: name of the metadata file
        """
        # Get the date adquired from the request json
        with rasterio.open(raster_path) as ds:
            bounds = ds.bounds
            dst_crs = "EPSG:4326"
            left, bottom, right, top = rasterio.warp.transform_bounds(
                ds.crs, dst_crs, *bounds
            )
            bbox = [left, bottom, right, top]

        payload_data = request_json['request']['payload']['input']['data'][0]
        # Get the data type from the request json
        data_type = payload_data['type']
        # Get the acquision date from the request json
        if 'timeRange' in payload_data['dataFilter']:
            data_acquisition_date = payload_data['dataFilter']['timeRange']['from']
        else:   # DEM data does not have a timeRange
            data_acquisition_date = None

        # Generate the metadata.json file
        metadata_path = join(output_folder, f"{output_name}.json")
        metadata = {
            "acquisition-date": data_acquisition_date,
            "bounding-box": bbox,
            "type": data_type,
        }

        with open(metadata_path, "w") as f:
            json.dump(metadata, f)

        return data_type



'''
Evalscripts for Sentinel Hub requests
'''


class EvalScripts:
    """
    Class that defines the needed Sentinel Hub evalscripts
    """

    SENTINEL_1 = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["VH", "VV"]
                        }],
                        output: {
                            id: "default",
                            bands: 2
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.VH, sample.VV];
                }
                """
    
    SENTINEL_2 = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["B01", 
                                    "B02", 
                                    "B03", 
                                    "B04",
                                    "B05", 
                                    "B06", 
                                    "B07", 
                                    "B08", 
                                    "B8A",
                                    "B09",
                                    "B11", 
                                    "B12"],
                            units: "DN"
                        }],
                        output: {
                            bands: 12,
                            sampleType: "INT16"
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.B01, 
                            sample.B02, 
                            sample.B03, 
                            sample.B04, 
                            sample.B05, 
                            sample.B06, 
                            sample.B07, 
                            sample.B08, 
                            sample.B8A,
                            sample.B09,
                            sample.B11, 
                            sample.B12];
                }
                """
    
    DEM = """
        //VERSION=3

        function setup() {
            return {
                input: ["DEM"],
                output: { id: "default",
                        bands: 1,
                        sampleType: SampleType.FLOAT32
                },
            }
        }

        function evaluatePixel(sample) {
            return [sample.DEM]
        }
        """



from pathlib import Path
from glob import glob
import yaml
from tqdm import tqdm
import os

from ..auth import with_auth
from .metadata import Metadata
from ..repos import DatasetsAPIRepo, FilesAPIRepo
from .utils import calculate_checksum


def ingest_dataset(path, verbose=False, logger=print):
    path = Path(path)
    if not path.is_dir():
        raise Exception("Path must be a folder")
    # if "catalog.json" in [f.name for f in path.iterdir()]:
    #     return ingest_stac(path / "catalog.json", logger)
    return ingest_folder(path, verbose, logger)


@with_auth
def ingest_folder(folder, verbose=False, logger=print, user=None):
    repo, files_repo = DatasetsAPIRepo(), FilesAPIRepo()
    logger(f"Uploading directory {folder}...")
    # get all files in directory recursively
    items = [Path(item) for item in glob(str(folder) + "/**/*", recursive=True)]
    # remove directories
    items = [item for item in items if not item.is_dir()]
    if len(items) == 0:
        raise Exception("No files found in directory")
    if not any(item.name == "metadata.yml" for item in items):
        raise Exception("metadata.yml not found in directory")
    # load metadata
    metadata = yaml.safe_load(open(folder.joinpath("metadata.yml"), "r").read()) or {}
    metadata = Metadata(**metadata)
    # remove metadata.yml from files
    items = [item for item in items if item.name != "metadata.yml"]
    # if zip or tar file, send error
    if any(item.suffix.endswith((".zip", ".tar", ".tar.gz", ".gz")) for item in items):
        raise Exception(
            f"At least one zip, tar or gz file found in {folder}, please unzip and try again"
        )
    # create dataset
    data, error = repo.create_dataset(metadata.dict(), user["id_token"])
    # dataset may already exist, and will return an error, but if user is owner continue ingesting files
    current_files = []
    if error:
        data, error2 = repo.retrieve_dataset(metadata.name)
        if error2:
            raise Exception(error)
        if data["uid"] != user["sub"]:
            raise Exception("Dataset already exists.")
        data["dataset_id"] = data["id"]
    dataset_id = data["dataset_id"]
    # create new version
    data, error = repo.create_version(dataset_id, user["id_token"])
    if error:
        raise Exception(error)
    version = data["version"]
    # upload files
    current_files = []
    if version > 1:
        current_files, error = files_repo.retrieve_dataset_files(
            dataset_id, version - 1
        )
        if error:
            # print("retreive dataset files error: ", error)
            current_files = []
    for item in tqdm(items, desc="Uploading files", unit="files", disable=verbose):
        data = ingest_file(
            str(item),
            dataset_id,
            version,
            str(item.relative_to(folder).parent),
            logger=logger,
            verbose=verbose,
            user=user,
            current_files=current_files,
        )
    return data


def ingest_file(
    file,
    dataset_id,
    version,
    parent,
    logger=None,
    verbose=True,
    root=None,
    user=None,
    current_files=[],
):
    id_token = user["id_token"]
    if verbose:
        logger(f"Uploading file {file}...")
    repo = FilesAPIRepo()
    if file.startswith("http://") or file.startswith("https://"):
        raise NotImplementedError("URL ingestion not implemented yet")
        # data, error = repo.ingest_file_url(file, dataset_id, id_token)
    else:
        file_path = Path(file)
        if not file_path.is_absolute():
            # file_path = glob(
            #     str(root) + "/**/" + os.path.basename(file_path),
            #     recursive=True,
            # )
            # if len(file_path) == 0:
            #     raise Exception(f"File {file} not found")
            # elif len(file_path) > 1:
            #     raise Exception(f"Multiple files found for {file}")
            # file_path = file_path[0]
            file_path = str(file_path.absolute())
        if verbose:
            logger("Computing checksum...")
        checksum = calculate_checksum(file_path)
        # check if file already exists in dataset
        filename = os.path.basename(file_path)
        if parent != ".":
            filename = parent + "/" + filename
        if len(current_files) > 0:
            matches = [
                f
                for f in current_files
                if f["filename"] == filename and f["checksum"] == checksum
            ]  # this could slow down ingestion in large datasets... should think of faster search algos, puede que sea mejor hacer el re-upload simplemente...
            if len(matches) == 1:
                if verbose:
                    print(f"File {file_path} already exists in dataset, skipping...")
                data, error = repo.ingest_existing_file(
                    filename,
                    dataset_id,
                    version,
                    matches[0]["version"],
                    id_token,
                    checksum,
                )
                if error:
                    raise Exception(error)
                if verbose:
                    logger("Done")
                return data
        if verbose:
            logger("Ingesting file...")
        filesize = os.path.getsize(file_path)
        # ingest small file
        if filesize < 1024 * 1024 * 16:  # 16 MB
            data, error = repo.ingest_file(
                file_path,
                dataset_id,
                version,
                parent,
                id_token,
                checksum,
            )
            if error:
                raise Exception(error)
            if verbose:
                logger("Done")
            return data
        raise NotImplementedError("Large file ingestion not implemented yet")
        # # ingest large file
        # upload_id, parts = repo.prepare_large_upload(
        #     file_path, dataset_id, checksum, id_token
        # )
        # repo.ingest_large_dataset(file_path, upload_id, id_token, parts)
        # if verbose:
        #     logger("\nCompleting upload...")
        # data, error = repo.complete_upload(id_token, upload_id)
    if error:
        raise Exception(error)
    if verbose:
        logger("Done")
    return data


# @with_auth
# def ingest_stac(stac_catalog, logger=None, user=None):
#     api_repo = APIRepo()
#     ingest = IngestSTAC(api_repo, ingest_file, logger)
#     inputs = ingest.Inputs(stac_catalog=stac_catalog, user=user)
#     outputs = ingest(inputs)
#     return outputs.dataset



import time

from ..repos import AuthRepo, AuthAPIRepo
from .errors import LoginError, AuthTimeOut
from .is_logged import is_logged


def auth(max_t=30, interval=2):
    user = is_logged()
    if user:
        return user
    repo, api_repo = AuthRepo(), AuthAPIRepo()
    response = api_repo.login()
    if response.status_code != 200:
        raise LoginError()
    data = response.json()
    print("On your computer or mobile device navigate to: ", data["login_url"])
    authenticated = False
    t0 = time.time()
    while not authenticated and time.time() - t0 < max_t:
        response = api_repo.token(data["code"])
        token_data = response.json()
        if response.status_code == 200:
            print("Authenticated!")
            print("- Id Token: {}...".format(token_data["id_token"][:10]))
            # get user credentials
            credentials = api_repo.retrieve_credentials(token_data["id_token"])[0]
            token_data.update(credentials)
            # save token data in file
            creds_path = repo.save_creds(token_data)
            print("Saved credentials to: ", creds_path)
            current_user = repo.decode_token(token_data)
            authenticated = True
            current_user["id_token"] = token_data["id_token"]
            return current_user
        else:
            time.sleep(interval)
    if not authenticated:
        raise AuthTimeOut()


# auth decorator
def with_auth(func):
    def wrapper(*args, **kwargs):
        user = auth()
        return func(*args, **kwargs, user=user)

    return wrapper



"""
Module for managing the Airbus configuration and data access
"""

import json
import requests
from requests.exceptions import ConnectTimeout, ReadTimeout
from urllib3.exceptions import TimeoutError
import time

from typing import Optional, Iterable
from os.path import join, exists
from .parameters import *
from .utils import get_airbus_access_token
from ...tools.tools import expand_time_interval, bbox_to_coordinates


class AirbusClient():
  """
  Client class to manage the Sentinel Hub Python interface.
  """

  def __init__(self, 
               access_token: str,
               api_key: str,
               ) -> None:
    """
    Constructor

    Params
    ----------
    access_token: str
        Access token
    api_key: str
        API key
    """
    self.airbus_access_token = access_token
    self._api_key = api_key

  def get_total_products_price(self,
                               payload: dict,
                               all_info: Optional[bool] = False
                               ) -> dict:
    """
    """
    if all_info:
      response = list()
    else:
      response = 0
    
    for location_id, location_data in payload.items():
      product_id = location_data['image']
      if id:
          price_response = self.get_product_price(product_id, location_data['bounding_box'])
          if all_info:
            response.append(price_response)
          else:
            response += price_response['price']['credits']

    return response

  def get_product_price(self,
                        product_id: str, 
                        coordinates: Iterable,
                        product_type: Optional[AirbusProductType] = AirbusProductType.MULTISPECTRAL,
                        image_format: Optional[AirbusImageFormat] = AirbusImageFormat.GEOTIFF,
                        processing: Optional[AirbusRadiometricProcessing] = AirbusRadiometricProcessing.REFLECTANCE
                        ) -> dict:
    """
    Get product price
    
    Params
    ----------
    product_id: str
        Product ID
    coordinates: tuple
        Polygon coordinates
    product_type: AirbusProductType
        Product type
    image_format: AirbusImageFormat
        Image format
    processing: AirbusRadiometricProcessing
        Radiometric processing
    Types are defined at parameters.py

    Returns
    ----------
    dict
        Product price
    """
    if (isinstance(coordinates, tuple) or isinstance(coordinates, list)) and len(coordinates) == 4:
      coordinates = bbox_to_coordinates(coordinates)

    headers = {
        'Authorization':  f'Bearer {self.airbus_access_token}',
        'Content-Type': "application/json",
        'Cache-Control': "no-cache",
        }
    
    # TODO make productType, imageformat and processing configurable with paramters controlled by classes like
    # TODO AirbusProductType.multiSpectral, AirbusImageFormat.geotiff, AirbusProcessing
    payload = {
      "kind": "order.data.product",
      "products": [
        {
          "productType": product_type,
          "radiometricProcessing": processing,
          "imageFormat": image_format,
          "crsCode": "urn:ogc:def:crs:EPSG::4326",
          "id": product_id,
          "aoi": {
            "type": "Polygon",
            "coordinates": [
              coordinates
            ]
          }
        }
      ]
    }
    response = requests.request("POST", AirbusURL.PRICES, json=payload, headers=headers)

    return response.json()
  
  def place_product_order(self,
                          product_id: str,
                          coordinates: Iterable,
                          product_type: Optional[AirbusProductType] = AirbusProductType.MULTISPECTRAL,
                          image_format: Optional[AirbusImageFormat] = AirbusImageFormat.GEOTIFF,
                          processing: Optional[AirbusRadiometricProcessing] = AirbusRadiometricProcessing.REFLECTANCE
                          ) -> dict:
    """
    Place product order
    
    Params
    ----------
    product_id: str
        Product ID
    bounding_box: tuple
        Bounding box
    product_type: AirbusProductType
        Product type
    image_format: AirbusImageFormat
        Image format
    processing: AirbusRadiometricProcessing
        Radiometric processing
    Types are defined at parameters.py

    Returns
    ----------
    dict
        Order data
    """
    if (isinstance(coordinates, tuple) or isinstance(coordinates, list)) and len(coordinates) == 4:
      coordinates = bbox_to_coordinates(coordinates)
    
    payload = {
      "kind": "order.data.product",
      "products": [
        {
          "productType": product_type,
          "radiometricProcessing": processing,
          "imageFormat": image_format,
          "crsCode": "urn:ogc:def:crs:EPSG::4326",
          "id": product_id,
          "aoi": {
            "type": "Polygon",
            "coordinates": [
              coordinates
            ]
          }
        }
      ]
    }

    headers = {
        'Authorization': f"Bearer {self.airbus_access_token}",
    }

    response = requests.request("POST", AirbusURL.ORDERS, json=payload, headers=headers)

    return response.json()
  
  def search_image(self,
                   bounding_box: Iterable,
                   acquisition_date: Iterable,
                   timeout: Optional[int] = 10
                   ) -> dict:
    """
    Search image

    Params
    ----------
    bounding_box: tuple or list
        Bounding box
    acquisition_date: tuple or list
        Acquisition date
    timeout: int
        Timeout
      
    Returns
    ----------
    dict
        Image data
    """
    if isinstance(acquisition_date, tuple) or isinstance(acquisition_date, list):
      acquisition_date = "[" + ",".join(acquisition_date) + "]"
    if (isinstance(bounding_box, tuple) or isinstance(bounding_box, list)) and len(bounding_box) == 4:
      bounding_box = ",".join(str(num) for num in bounding_box)

    querystring = {"acquisitionDate": str(acquisition_date),
                   "bbox": bounding_box
                   }

    headers = {
        'authorization': f"Bearer {self.airbus_access_token}",
        'cache-control': "no-cache",
        }

    try:
      response = requests.request("GET", AirbusURL.SEARCH, headers=headers, params=querystring, verify=False, timeout=timeout)
      return response.json()
    except json.decoder.JSONDecodeError:
      print('JSONDecodeError')
      print(response)
    except ReadTimeout:
      print("ReadTimeout")
      print(response)

  def search_images_close_in_time(self,
                                  payload_dict: dict,
                                  path: Optional[str] = None,
                                  max_days: Optional[int] = 30
                                  ) -> dict:
    """
    Search images close in time

    Params
    ----------
    payload_dict: dict
        Payload dictionary
    max_days: int
        Maximum days to search

    Returns
    ----------
    dict
        Dictionary with the image data, as {location_id: image_data}, to
        maintain track of the location
    """
    responses = dict()
    if path:
      responses_path = join(path, 'airbus_images_response.json')
      if exists(responses_path):
        with open(responses_path, 'r') as f:
          responses = json.load(f)

    for location_id, location_info in list(payload_dict.items()):
        bounding_box, time_interval = location_info['bounding_box'], location_info['time_interval']
        days = 1

        if location_id in responses:
          continue

        while days <= max_days:
            try:
              response = self.search_image(bounding_box, time_interval)
            except ConnectTimeout or TimeoutError:
              # Wait 5 seconds and try again
              time.sleep(5)
              # Restart the connection
              self.airbus_access_token = get_airbus_access_token(self._api_key)
              # Continue with the search
              response = self.search_image(bounding_box, time_interval)

            total_results = response['totalResults']

            if total_results > 0:
                # By default, the search results are sorted per acquisition date 
                # (newest data is displayed first) and per cloud coverage (less cloudy images are displayed first).
                # So, we can stop the search when we find the first image
                if total_results > 1:
                    response['features'] = [response['features'][0]]
                responses[location_id] = response['features'][0]
                if path:
                  with open(responses_path, 'w') as f:
                    json.dump(responses, f)
                break

            time_interval = expand_time_interval(time_interval)
            days += 1

        # If no image is found, we add a None value to the dictionary
        if total_results == 0:
          responses[location_id] = None
          if path:
            with open(responses_path, 'w') as f:
              json.dump(responses, f)

    return responses

  def format_product_payload(self,
                             location_payload: dict,
                             images_response: dict
                             ) -> dict:
    """
    """
    # TODO put in airbus.tools?
    for id, info in location_payload.items():
      # Add new key to the dictionary
      location_payload[id]['image_response'] = images_response[id] if id in images_response else None

    return location_payload

  def split_product_payload(self,
                            product_payload: dict
                            ) -> dict:
    """
    """
    # TODO put in airbus.tools?
    # split the product payload depending on if 'image' is None or not
    product_payload_with_image = dict()
    product_payload_without_image = dict()

    for id, info in product_payload.items():
      if info['image']:
        product_payload_with_image[id] = info
      else:
        product_payload_without_image[id] = info
    
    return product_payload_with_image, product_payload_without_image

  def get_all_order_status(self):
    """
    """
    headers = {
      'Authorization': f"Bearer {self.airbus_access_token}",
      'Content-Type': "application/json",
      'Cache-Control': "no-cache",
    }

    response = requests.request("GET", AirbusURL.ALL_ORDERS_STATUS, headers=headers)

    return response.text

  def get_account_information(self):
    """
    """
    headers = {
      'Authorization': f"Bearer {self.airbus_access_token}",
      'Content-Type': "application/json",
      'Cache-Control': "no-cache",
    }

    response = requests.request("GET", AirbusURL.ACCOUNT, headers=headers)

    return response.json()

  def get_user_roles(self):
    """
    """
    headers = {
      'Authorization': f"Bearer {self.airbus_access_token}"
    }

    response = requests.request("GET", AirbusURL.ROLES, headers=headers)

    return response.json()
